###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module that defines the configuration file of an AiiDA instance and functions to create and load it.

Despite the import of the annotations backport below which enables postponed type annotation evaluation as implemented
with PEP 563 (https://peps.python.org/pep-0563/), this is not compatible with ``pydantic`` for Python 3.9 and older (
See https://github.com/pydantic/pydantic/issues/2678 for details).
"""

from __future__ import annotations

import codecs
import contextlib
import io
import json
import os
import uuid
from typing import Any, Dict, List, Optional, Tuple

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    field_serializer,
    field_validator,
)

from aiida.common.exceptions import ConfigurationError, EntryPointError, StorageMigrationError
from aiida.common.log import AIIDA_LOGGER, LogLevels

from .options import Option, get_option, get_option_names, parse_option
from .profile import Profile

LOGGER = AIIDA_LOGGER.getChild('manage.configuration.config')


class ConfigVersionSchema(BaseModel, defer_build=True):
    """Schema for the version configuration of an AiiDA instance."""

    CURRENT: int
    OLDEST_COMPATIBLE: int


class ProfileOptionsSchema(BaseModel, defer_build=True):
    """Schema for the options of an AiiDA profile."""

    model_config = ConfigDict(use_enum_values=True)

    runner__poll__interval: int = Field(60, description='Polling interval in seconds to be used by process runners.')
    daemon__default_workers: int = Field(
        1, description='Default number of workers to be launched by `verdi daemon start`.'
    )
    daemon__timeout: int = Field(
        2,
        description='Used to set default timeout in the `DaemonClient` for calls to the daemon.',
    )
    daemon__worker_process_slots: int = Field(
        200, description='Maximum number of concurrent process tasks that each daemon worker can handle.'
    )
    daemon__recursion_limit: int = Field(3000, description='Maximum recursion depth for the daemon workers.')
    db__batch_size: int = Field(
        100000,
        description='Batch size for bulk CREATE operations in the database. Avoids hitting MaxAllocSize of PostgreSQL '
        '(1GB) when creating large numbers of database records in one go.',
    )
    verdi__shell__auto_import: str = Field(
        ':',
        description='Additional modules/functions/classes to be automatically loaded in `verdi shell`, split by `:`.',
    )
    logging__aiida_loglevel: LogLevels = Field(
        'REPORT', description='Minimum level to log to daemon log and the `DbLog` table for the `aiida` logger.'
    )
    logging__verdi_loglevel: LogLevels = Field(
        'REPORT', description='Minimum level to log to console when running a `verdi` command.'
    )
    logging__disk_objectstore_loglevel: LogLevels = Field(
        'INFO', description='Minimum level to log to daemon log and the `DbLog` table for `disk_objectstore` logger.'
    )
    logging__db_loglevel: LogLevels = Field('REPORT', description='Minimum level to log to the DbLog table.')
    logging__plumpy_loglevel: LogLevels = Field(
        'WARNING', description='Minimum level to log to daemon log and the `DbLog` table for the `plumpy` logger.'
    )
    logging__kiwipy_loglevel: LogLevels = Field(
        'WARNING', description='Minimum level to log to daemon log and the `DbLog` table for the `kiwipy` logger'
    )
    logging__paramiko_loglevel: LogLevels = Field(
        'WARNING', description='Minimum level to log to daemon log and the `DbLog` table for the `paramiko` logger'
    )
    logging__alembic_loglevel: LogLevels = Field(
        'WARNING', description='Minimum level to log to daemon log and the `DbLog` table for the `alembic` logger'
    )
    logging__sqlalchemy_loglevel: LogLevels = Field(
        'WARNING', description='Minimum level to log to daemon log and the `DbLog` table for the `sqlalchemy` logger'
    )
    logging__circus_loglevel: LogLevels = Field(
        'INFO', description='Minimum level to log to daemon log and the `DbLog` table for the `circus` logger'
    )
    logging__aiopika_loglevel: LogLevels = Field(
        'WARNING', description='Minimum level to log to daemon log and the `DbLog` table for the `aiopika` logger'
    )
    warnings__showdeprecations: bool = Field(True, description='Whether to print AiiDA deprecation warnings.')
    warnings__rabbitmq_version: bool = Field(
        True, description='Whether to print a warning when an incompatible version of RabbitMQ is configured.'
    )
    transport__task_retry_initial_interval: int = Field(
        20, description='Initial time interval for the exponential backoff mechanism.'
    )
    transport__task_maximum_attempts: int = Field(
        5, description='Maximum number of transport task attempts before a Process is Paused.'
    )
    rmq__task_timeout: int = Field(10, description='Timeout in seconds for communications with RabbitMQ.')
    storage__sandbox: Optional[str] = Field(
        None, description='Absolute path to the directory to store sandbox folders.'
    )
    caching__default_enabled: bool = Field(False, description='Enable calculation caching by default.')
    caching__enabled_for: List[str] = Field([], description='Calculation entry points to enable caching on.')
    caching__disabled_for: List[str] = Field([], description='Calculation entry points to disable caching on.')

    @field_validator('caching__enabled_for', 'caching__disabled_for')
    @classmethod
    def validate_caching_identifier_pattern(cls, value: List[str]) -> List[str]:
        """Validate the caching identifier patterns."""
        from aiida.manage.caching import _validate_identifier_pattern

        for identifier in value:
            _validate_identifier_pattern(identifier=identifier, strict=True)
        return value


class GlobalOptionsSchema(ProfileOptionsSchema, defer_build=True):
    """Schema for the global options of an AiiDA instance."""

    autofill__user__email: Optional[str] = Field(
        None, description='Default user email to use when creating new profiles.'
    )
    autofill__user__first_name: Optional[str] = Field(
        None, description='Default user first name to use when creating new profiles.'
    )
    autofill__user__last_name: Optional[str] = Field(
        None, description='Default user last name to use when creating new profiles.'
    )
    autofill__user__institution: Optional[str] = Field(
        None, description='Default user institution to use when creating new profiles.'
    )
    rest_api__profile_switching: bool = Field(
        False, description='Toggle whether the profile can be specified in requests submitted to the REST API.'
    )
    warnings__development_version: bool = Field(
        True,
        description='Whether to print a warning when a profile is loaded while a development version is installed.',
    )


class ProfileStorageConfig(BaseModel, defer_build=True):
    """Schema for the storage backend configuration of an AiiDA profile."""

    backend: str
    config: Dict[str, Any]


class ProcessControlConfig(BaseModel, defer_build=True):
    """Schema for the process control configuration of an AiiDA profile."""

    broker_protocol: str = Field('amqp', description='Protocol for connecting to the message broker.')
    broker_username: str = Field('guest', description='Username for message broker authentication.')
    broker_password: str = Field('guest', description='Password for message broker.')
    broker_host: str = Field('127.0.0.1', description='Hostname of the message broker.')
    broker_port: int = Field(5432, description='Port of the message broker.')
    broker_virtual_host: str = Field('', description='Virtual host to use for the message broker.')
    broker_parameters: dict[str, Any] = Field(
        default_factory=dict, description='Arguments to be encoded as query parameters.'
    )


class ProfileSchema(BaseModel, defer_build=True):
    """Schema for the configuration of an AiiDA profile."""

    uuid: str = Field(description='A UUID that uniquely identifies the profile.', default_factory=uuid.uuid4)
    storage: ProfileStorageConfig
    process_control: ProcessControlConfig
    default_user_email: Optional[str] = None
    test_profile: bool = False
    options: Optional[ProfileOptionsSchema] = None

    @field_serializer('uuid')
    def serialize_dt(self, value: uuid.UUID, _info):
        return str(value)


class ConfigSchema(BaseModel, defer_build=True):
    """Schema for the configuration of an AiiDA instance."""

    CONFIG_VERSION: Optional[ConfigVersionSchema] = None
    profiles: Optional[dict[str, ProfileSchema]] = None
    options: Optional[GlobalOptionsSchema] = None
    default_profile: Optional[str] = None


class Config:
    """Object that represents the configuration file of an AiiDA instance."""

    KEY_VERSION = 'CONFIG_VERSION'
    KEY_VERSION_CURRENT = 'CURRENT'
    KEY_VERSION_OLDEST_COMPATIBLE = 'OLDEST_COMPATIBLE'
    KEY_DEFAULT_PROFILE = 'default_profile'
    KEY_PROFILES = 'profiles'
    KEY_OPTIONS = 'options'
    KEY_SCHEMA = '$schema'

    @classmethod
    def from_file(cls, filepath):
        """Instantiate a configuration object from the contents of a given file.

        .. note:: if the filepath does not exist an empty file will be created with the current default configuration
            and will be written to disk. If the filepath does already exist but contains a configuration with an
            outdated schema, the content will be migrated and then written to disk.

        :param filepath: the absolute path to the configuration file
        :return: `Config` instance
        """
        from aiida.cmdline.utils import echo

        from .migrations import check_and_migrate_config, config_needs_migrating

        try:
            with open(filepath, 'rb') as handle:
                config = json.load(handle)
        except FileNotFoundError:
            config = Config(filepath, check_and_migrate_config({}))
            config.store()
        else:
            migrated = False

            # If the configuration file needs to be migrated first create a specific backup so it can easily be reverted
            if config_needs_migrating(config, filepath):
                migrated = True
                echo.echo_warning(f'current configuration file `{filepath}` is outdated and will be migrated')
                filepath_backup = cls._backup(filepath)
                echo.echo_warning(f'original backed up to `{filepath_backup}`')

            config = Config(filepath, check_and_migrate_config(config))

            if migrated:
                config.store()

        return config

    @classmethod
    def _backup(cls, filepath):
        """Create a backup of the configuration file with the given filepath.

        :param filepath: absolute path to the configuration file to backup
        :return: the absolute path of the created backup
        """
        import shutil

        from aiida.common import timezone

        filepath_backup = None

        # Keep generating a new backup filename based on the current time until it does not exist
        while not filepath_backup or os.path.isfile(filepath_backup):
            filepath_backup = f"{filepath}.{timezone.now().strftime('%Y%m%d-%H%M%S.%f')}"

        shutil.copy(filepath, filepath_backup)

        return filepath_backup

    @staticmethod
    def validate(config: dict, filepath: Optional[str] = None):
        """Validate a configuration dictionary."""
        try:
            ConfigSchema(**config)
        except ValidationError as exception:
            raise ConfigurationError(f'invalid config schema: {filepath}: {exception!s}')

    def __init__(self, filepath: str, config: dict, validate: bool = True):
        """Instantiate a configuration object from a configuration dictionary and its filepath.

        If an empty dictionary is passed, the constructor will create the skeleton configuration dictionary.

        :param filepath: the absolute filepath of the configuration file
        :param config: the content of the configuration file in dictionary form
        :param validate: validate the dictionary against the schema
        """
        from .migrations import CURRENT_CONFIG_VERSION, OLDEST_COMPATIBLE_CONFIG_VERSION

        if validate:
            self.validate(config, filepath)

        self._filepath = filepath
        self._schema = config.get(self.KEY_SCHEMA, None)
        version = config.get(self.KEY_VERSION, {})
        self._current_version = version.get(self.KEY_VERSION_CURRENT, CURRENT_CONFIG_VERSION)
        self._oldest_compatible_version = version.get(
            self.KEY_VERSION_OLDEST_COMPATIBLE, OLDEST_COMPATIBLE_CONFIG_VERSION
        )
        self._profiles = {}

        known_keys = [self.KEY_SCHEMA, self.KEY_VERSION, self.KEY_PROFILES, self.KEY_OPTIONS, self.KEY_DEFAULT_PROFILE]
        unknown_keys = set(config.keys()) - set(known_keys)

        if unknown_keys:
            keys = ', '.join(unknown_keys)
            self.handle_invalid(f'encountered unknown keys [{keys}] in `{filepath}` which have been removed')

        try:
            self._options = config[self.KEY_OPTIONS]
        except KeyError:
            self._options = {}

        try:
            self._default_profile = config[self.KEY_DEFAULT_PROFILE]
        except KeyError:
            self._default_profile = None

        for name, config_profile in config.get(self.KEY_PROFILES, {}).items():
            self._profiles[name] = Profile(name, config_profile)

    def __eq__(self, other):
        """Two configurations are considered equal, when their dictionaries are equal."""
        return self.dictionary == other.dictionary

    def __ne__(self, other):
        """Two configurations are considered unequal, when their dictionaries are unequal."""
        return self.dictionary != other.dictionary

    def handle_invalid(self, message):
        """Handle an incoming invalid configuration dictionary.

        The current content of the configuration file will be written to a backup file.

        :param message: a string message to echo with describing the infraction
        """
        from aiida.cmdline.utils import echo

        filepath_backup = self._backup(self.filepath)
        echo.echo_warning(message)
        echo.echo_warning(f'backup of the original config file written to: `{filepath_backup}`')

    @property
    def dictionary(self) -> dict:
        """Return the dictionary representation of the config as it would be written to file.

        :return: dictionary representation of config as it should be written to file
        """
        config = {}
        if self._schema:
            config[self.KEY_SCHEMA] = self._schema

        config[self.KEY_VERSION] = self.version_settings
        config[self.KEY_PROFILES] = {name: profile.dictionary for name, profile in self._profiles.items()}

        if self._default_profile:
            config[self.KEY_DEFAULT_PROFILE] = self._default_profile

        if self._options:
            config[self.KEY_OPTIONS] = self._options

        return config

    @property
    def version(self):
        return self._current_version

    @version.setter
    def version(self, version):
        self._current_version = version

    @property
    def version_oldest_compatible(self):
        return self._oldest_compatible_version

    @version_oldest_compatible.setter
    def version_oldest_compatible(self, version_oldest_compatible):
        self._oldest_compatible_version = version_oldest_compatible

    @property
    def version_settings(self):
        return {
            self.KEY_VERSION_CURRENT: self.version,
            self.KEY_VERSION_OLDEST_COMPATIBLE: self.version_oldest_compatible,
        }

    @property
    def filepath(self):
        return self._filepath

    @property
    def dirpath(self):
        return os.path.dirname(self.filepath)

    @property
    def default_profile_name(self):
        """Return the default profile name.

        :return: the default profile name or None if not defined
        """
        return self._default_profile

    @property
    def profile_names(self):
        """Return the list of profile names.

        :return: list of profile names
        """
        return list(self._profiles.keys())

    @property
    def profiles(self):
        """Return the list of profiles.

        :return: the profiles
        :rtype: list of `Profile` instances
        """
        return list(self._profiles.values())

    def validate_profile(self, name):
        """Validate that a profile exists.

        :param name: name of the profile:
        :raises aiida.common.ProfileConfigurationError: if the name is not found in the configuration file
        """
        from aiida.common import exceptions

        if name not in self.profile_names:
            raise exceptions.ProfileConfigurationError(f'profile `{name}` does not exist')

    def get_profile(self, name: Optional[str] = None) -> Profile:
        """Return the profile for the given name or the default one if not specified.

        :return: the profile instance or None if it does not exist
        :raises aiida.common.ProfileConfigurationError: if the name is not found in the configuration file
        """
        from aiida.common import exceptions

        if not name and not self.default_profile_name:
            raise exceptions.ProfileConfigurationError(
                f'no default profile defined: {self._default_profile}\n{self.dictionary}'
            )

        if not name:
            name = self.default_profile_name

        self.validate_profile(name)

        return self._profiles[name]

    def create_profile(
        self,
        name: str,
        storage_backend: str,
        storage_config: dict[str, Any],
        broker_backend: str | None = None,
        broker_config: dict[str, Any] | None = None,
        is_test_profile: bool = False,
    ) -> Profile:
        """Create a new profile and initialise its storage.

        :param name: The profile name.
        :param storage_backend: The entry point to the :class:`aiida.orm.implementation.storage_backend.StorageBackend`
            implementation to use for the storage.
        :param storage_config: The configuration necessary to initialise and connect to the storage backend.
        :param broker_backend: The entry point to the :class:`aiida.brokers.Broker` implementation to use for the
            message broker.
        :param broker_config: The configuration necessary to initialise and connect to the broker.
        :returns: The created profile.
        :raises ValueError: If the profile already exists.
        :raises TypeError: If the ``storage_backend`` is not a subclass of
            :class:`aiida.orm.implementation.storage_backend.StorageBackend`.
        :raises EntryPointError: If the ``storage_backend`` does not have an associated entry point.
        :raises StorageMigrationError: If the storage cannot be initialised.
        """
        from aiida.brokers import Broker
        from aiida.orm.implementation.storage_backend import StorageBackend
        from aiida.plugins.entry_point import load_entry_point

        if name in self.profile_names:
            raise ValueError(f'The profile `{name}` already exists.')

        try:
            storage_cls = load_entry_point('aiida.storage', storage_backend)
        except EntryPointError as exception:
            raise ValueError(f'The entry point `{storage_backend}` could not be loaded.') from exception
        else:
            if not issubclass(storage_cls, StorageBackend):
                raise TypeError(
                    f'The `storage_backend={storage_backend}` is not a subclass of '
                    '`aiida.orm.implementation.storage_backend.StorageBackend`.'
                )

        storage_config = storage_cls.Model(**(storage_config or {})).model_dump()

        if broker_backend is not None:
            try:
                broker_cls = load_entry_point('aiida.brokers', broker_backend)
            except EntryPointError as exception:
                raise ValueError(f'The entry point `{broker_backend}` could not be loaded.') from exception
            else:
                if not issubclass(broker_cls, Broker):
                    raise TypeError(
                        f'The `broker_backend={broker_backend}` is not a subclass of `aiida.brokers.broker.Broker`.'
                    )

        profile = Profile(
            name,
            {
                'storage': {
                    'backend': storage_backend,
                    'config': storage_config,
                },
                'process_control': {
                    'backend': broker_backend,
                    'config': broker_config,
                },
                'test_profile': is_test_profile,
            },
        )

        LOGGER.report('Initialising the storage backend.')
        try:
            with contextlib.redirect_stdout(io.StringIO()):
                profile.storage_cls.initialise(profile)
        except Exception as exception:
            raise StorageMigrationError(
                f'Storage backend initialisation failed, probably because the configuration is incorrect:\n{exception}'
            )
        LOGGER.report('Storage initialisation completed.')

        self.add_profile(profile)
        self.store()

        return profile

    def add_profile(self, profile):
        """Add a profile to the configuration.

        :param profile: the profile configuration dictionary
        :return: self
        """
        self._profiles[profile.name] = profile
        return self

    def update_profile(self, profile):
        """Update a profile in the configuration.

        :param profile: the profile instance to update
        :return: self
        """
        self._profiles[profile.name] = profile
        return self

    def remove_profile(self, name):
        """Remove a profile from the configuration.

        :param name: the name of the profile to remove
        :raises aiida.common.ProfileConfigurationError: if the given profile does not exist
        :return: self
        """
        self.validate_profile(name)
        self._profiles.pop(name)
        return self

    def delete_profile(self, name: str, delete_storage: bool = True) -> None:
        """Delete a profile including its storage.

        :param delete_storage: Whether to delete the storage with all its data or not.
        """
        from aiida.plugins import StorageFactory

        profile = self.get_profile(name)
        is_default_profile: bool = profile.name == self.default_profile_name

        if delete_storage:
            storage_cls = StorageFactory(profile.storage_backend)
            storage = storage_cls(profile)
            storage.delete()
            LOGGER.report(f'Data storage deleted, configuration was: {profile.storage_config}')
        else:
            LOGGER.report(f'Data storage not deleted, configuration is: {profile.storage_config}')

        self.remove_profile(name)

        if is_default_profile and not self.profile_names:
            LOGGER.warning(f'`{name}` was the default profile, no profiles remain to set as default.')
            self.store()
            return

        if is_default_profile:
            LOGGER.warning(f'`{name}` was the default profile, setting `{self.profile_names[0]}` as the new default.')
            self.set_default_profile(self.profile_names[0], overwrite=True)

        self.store()

    def set_default_profile(self, name, overwrite=False):
        """Set the given profile as the new default.

        :param name: name of the profile to set as new default
        :param overwrite: when True, set the profile as the new default even if a default profile is already defined
        :raises aiida.common.ProfileConfigurationError: if the given profile does not exist
        :return: self
        """
        if self.default_profile_name and not overwrite:
            return self

        self.validate_profile(name)
        self._default_profile = name
        return self

    def set_default_user_email(self, profile: Profile, user_email: str) -> None:
        """Set the default user for the given profile.

        .. warning::

            This does not update the cached default user on the storage backend associated with the profile. To do so,
            use :meth:`aiida.manage.manager.Manager.set_default_user_email` instead.

        :param profile: The profile to update.
        :param user_email: The email of the user to set as the default user.
        """
        profile.default_user_email = user_email
        self.update_profile(profile)
        self.store()

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, value):
        self._options = value

    def set_option(self, option_name, option_value, scope=None, override=True):
        """Set a configuration option for a certain scope.

        :param option_name: the name of the configuration option
        :param option_value: the option value
        :param scope: set the option for this profile or globally if not specified
        :param override: boolean, if False, will not override the option if it already exists

        :returns: the parsed value (potentially cast to a valid type)
        """
        option, parsed_value = parse_option(option_name, option_value)

        if parsed_value is not None:
            value = parsed_value
        elif option.default is not None:
            value = option.default
        else:
            return

        if not option.global_only and scope is not None:
            self.get_profile(scope).set_option(option.name, value, override=override)
        elif option.name not in self.options or override:
            self.options[option.name] = value

        return value

    def unset_option(self, option_name: str, scope=None):
        """Unset a configuration option for a certain scope.

        :param option_name: the name of the configuration option
        :param scope: unset the option for this profile or globally if not specified
        """
        option = get_option(option_name)

        if scope is not None:
            self.get_profile(scope).unset_option(option.name)
        else:
            self.options.pop(option.name, None)

    def get_option(self, option_name, scope=None, default=True):
        """Get a configuration option for a certain scope.

        :param option_name: the name of the configuration option
        :param scope: get the option for this profile or globally if not specified
        :param default: boolean, If True will return the option default, even if not defined within the given scope
        :return: the option value or None if not set for the given scope
        """
        option = get_option(option_name)
        default_value = option.default if default else None

        if scope is not None:
            value = self.get_profile(scope).get_option(option.name, default_value)
        else:
            value = self.options.get(option.name, default_value)

        return value

    def get_options(self, scope: Optional[str] = None) -> Dict[str, Tuple[Option, str, Any]]:
        """Return a dictionary of all option values and their source ('profile', 'global', or 'default').

        :param scope: the profile name or globally if not specified
        :returns: (option, source, value)
        """
        profile = self.get_profile(scope) if scope else None
        output = {}
        for name in get_option_names():
            option = get_option(name)
            if profile and name in profile.options:
                value = profile.options.get(name)
                source = 'profile'
            elif name in self.options:
                value = self.options.get(name)
                source = 'global'
            elif option.default is not None:
                value = option.default
                source = 'default'
            else:
                continue
            output[name] = (option, source, value)
        return output

    def store(self):
        """Write the current config to file.

        .. note:: if the configuration file already exists on disk and its contents differ from those in memory, a
            backup of the original file on disk will be created before overwriting it.

        :return: self
        """
        import tempfile

        from aiida.common.files import md5_file, md5_from_filelike

        from .settings import DEFAULT_CONFIG_INDENT_SIZE

        # If the filepath of this configuration does not yet exist, simply write it.
        if not os.path.isfile(self.filepath):
            self._atomic_write()
            return self

        # Otherwise, we write the content to a temporary file and compare its md5 checksum with the current config on
        # disk. When the checksums differ, we first create a backup and only then overwrite the existing file.
        with tempfile.NamedTemporaryFile() as handle:
            json.dump(self.dictionary, codecs.getwriter('utf-8')(handle), indent=DEFAULT_CONFIG_INDENT_SIZE)
            handle.seek(0)

            if md5_from_filelike(handle) != md5_file(self.filepath):
                self._backup(self.filepath)

        self._atomic_write()

        return self

    def _atomic_write(self, filepath=None):
        """Write the config as it is in memory, i.e. the contents of ``self.dictionary``, to disk.

        .. note:: this command will write the config from memory to a temporary file in the same directory as the
            target file ``filepath``. It will then use ``os.rename`` to move the temporary file to ``filepath`` which
            will be overwritten if it already exists. The ``os.rename`` is the operation that gives the best guarantee
            of being atomic within the limitations of the application.

        :param filepath: optional filepath to write the contents to, if not specified, the default filename is used.
        """
        import tempfile

        from .settings import DEFAULT_CONFIG_INDENT_SIZE, DEFAULT_UMASK

        umask = os.umask(DEFAULT_UMASK)

        if filepath is None:
            filepath = self.filepath

        # Create a temporary file in the same directory as the target filepath, which guarantees that the temporary
        # file is on the same filesystem, which is necessary to be able to use ``os.rename``. Since we are moving the
        # temporary file, we should also tell the tempfile to not be automatically deleted as that will raise.
        with tempfile.NamedTemporaryFile(dir=os.path.dirname(filepath), delete=False, mode='w') as handle:
            try:
                json.dump(self.dictionary, handle, indent=DEFAULT_CONFIG_INDENT_SIZE)
            finally:
                os.umask(umask)

            handle.flush()
            os.rename(handle.name, self.filepath)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Functions to dynamically generate a WorkChain from a reversed polish notation expression."""

import collections
import hashlib
import os
from pathlib import Path
from string import Template

from .expression import OPERATORS

INDENTATION_WIDTH = 4

BLOCK_TPL_INI = '{indent:}cls.setup,\n'
BLOCK_TPL_ADD_BASE = """
    {indent:}if_(cls.is_positive)(
    {indent_inner:}cls.add
    {indent:}).else_(
    {indent_inner:}cls.subtract
    {indent:}),
    """

BLOCK_TPL_ADD_WORK = """
    {indent:}if_(cls.is_positive)(
    {indent_inner:}cls.add
    {indent:}).else_(
    {indent_inner:}cls.subtract_calcfunction
    {indent:}),
    """

BLOCK_TPL_ADD_CALC = """
    {indent:}if_(cls.is_positive)(
    {indent_inner:}cls.add_calculation,
    {indent_inner:}cls.post_add,
    {indent:}).else_(
    {indent_inner:}cls.subtract
    {indent:}),
    """

BLOCK_TPL_ADD_BOTH = """
    {indent:}if_(cls.is_positive)(
    {indent_inner:}cls.add_calculation,
    {indent_inner:}cls.post_add,
    {indent:}).else_(
    {indent_inner:}cls.subtract_calcfunction
    {indent:}),
    """

BLOCK_TPL_MUL = """
    {indent:}cls.pre_iterate,
    {indent:}while_(cls.iterate)(
    {block:}{indent:}),
    """

BLOCK_TPL_POW = """
    {indent:}cls.raise_power,
    {indent:}cls.post_raise_power,
    """

BLOCK_TPL_END = """
    {indent:}cls.results
    """


def generate_outlines(expression):
    """For a given expression in Reverse Polish Notation, generate the nested symbolic structure of the outlines.

    :param expression: a valid expression
    :return: a nested list structure of strings representing the structure of the outlines
    """
    stack = collections.deque()
    values = []
    outline = [['add']]

    for part in expression.split():
        if part not in OPERATORS:
            stack.appendleft(part)
            values.append(part)
        else:
            stack.popleft()

            sub_outline = outline[-1]

            if part == '+':
                sub_outline.append('add')
            elif part == '*':
                sub_outline = outline.pop()
                outline.append([sub_outline])
            elif part == '^':
                outline.append(['pow'])

    for sub_outline in outline:
        sub_outline.append('ini')
        sub_outline.append('end')

    return outline, values


def format_outlines(outlines, use_calculations=False, use_calcfunctions=False):
    """Given the symbolic structure of the workchain outlines produced by ``generate_outlines``, format the actual
    string form of those workchain outlines

    :param outlines: the list of symbolic outline structures
    :param use_calculations: use CalcJobs for the add operations
    :param use_calcfunctions: use calcfunctions for the subtract operations
    :return: a list of outline strings
    """
    outline_strings = []

    for sub_outline in outlines:
        outline_string = ''

        for instruction in sub_outline:
            if instruction == 'ini':
                outline_string = BLOCK_TPL_INI.format(indent=format_indent()) + outline_string
            elif instruction == 'add':
                outline_string = format_block(instruction, 0, use_calculations, use_calcfunctions) + outline_string
            elif instruction == 'pow':
                outline_string += BLOCK_TPL_POW.format(indent=format_indent())
            elif instruction == 'end':
                outline_string += BLOCK_TPL_END.format(indent=format_indent())
            else:
                outline_string += format_block(instruction, 0, use_calculations, use_calcfunctions)

        outline_strings.append(outline_string)

    return outline_strings


def format_block(instruction, level=0, use_calculations=False, use_calcfunctions=False):
    """Format the instruction into its proper string form

    :param use_calculations: use CalcJobs for the add operations
    :param use_calcfunctions: use calcfunctions for the subtract operations
    :return: the string representation of the instruction
    """
    block = ''
    string = ''

    if isinstance(instruction, list):
        for sub_instruction in instruction:
            if sub_instruction == 'add':
                block = format_block(sub_instruction, level + 1, use_calculations, use_calcfunctions) + block
            else:
                block += format_block(sub_instruction, level + 1, use_calculations, use_calcfunctions)

        string += BLOCK_TPL_MUL.format(indent=format_indent(level), level=level, block=block)

    elif instruction == 'pow':
        string += BLOCK_TPL_POW.format(indent=format_indent(level))

    elif instruction == 'add':
        if use_calculations and use_calcfunctions:
            string = BLOCK_TPL_ADD_BOTH.format(indent=format_indent(level), indent_inner=format_indent(level + 1))
        elif use_calculations:
            string = BLOCK_TPL_ADD_CALC.format(indent=format_indent(level), indent_inner=format_indent(level + 1))
        elif use_calcfunctions:
            string = BLOCK_TPL_ADD_WORK.format(indent=format_indent(level), indent_inner=format_indent(level + 1))
        else:
            string = BLOCK_TPL_ADD_BASE.format(indent=format_indent(level), indent_inner=format_indent(level + 1))

    return string


def format_indent(level=0, width=INDENTATION_WIDTH):
    """Format the indentation for the given indentation level and indentation width

    :param level: the level of indentation
    :param width: the width in spaces of a single indentation
    :return: the indentation string
    """
    return ' ' * level * width


def write_workchain(outlines, directory=None) -> Path:
    """Given a list of string formatted outlines, write the corresponding workchains to file

    :returns: file path
    """
    dirpath = os.path.dirname(os.path.realpath(__file__))
    template_dir = os.path.join(dirpath, 'template')
    template_file_base = os.path.join(template_dir, 'base.tpl')
    template_file_workchain = os.path.join(template_dir, 'workchain.tpl')

    if directory is None:
        directory = os.path.join(dirpath, os.path.pardir, 'polish_workchains')

    directory = Path(directory)

    directory.mkdir(parents=True, exist_ok=True)
    (directory / '__init__.py').touch()

    with open(template_file_base, 'r', encoding='utf8') as handle:
        template_base = handle.readlines()

    with open(template_file_workchain, 'r', encoding='utf8') as handle:
        template_workchain = Template(handle.read())

    code_strings = []

    for line in template_base:
        code_strings.append(line)
    code_strings.append('\n')

    counter = len(outlines) - 1
    for outline in outlines:
        outline_string = ''
        for subline in outline.split('\n'):
            outline_string += f'\t\t\t{subline}\n'

        if counter == len(outlines) - 1:
            child_class = None
        else:
            child_class = f'Polish{counter + 1:02d}WorkChain'

        subs = {
            'class_name': f'Polish{counter:02d}WorkChain',
            'child_class': child_class,
            'outline': outline_string,
        }
        code_strings.append(template_workchain.substitute(**subs))
        code_strings.append('\n\n')

        counter -= 1

    code_string = '\n'.join(code_strings)
    hashed = hashlib.md5(code_string.encode('utf8')).hexdigest()

    filepath = directory / f'polish_{hashed}.py'

    filepath.write_text(code_string)

    return filepath



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tools to operate on AiiDA ORM class instances

What functionality should go directly in the ORM class in `aiida.orm` and what in `aiida.tools`?

    - The ORM class should define basic functions to set and get data from the object
    - More advanced functionality to operate on the ORM class instances can be placed in `aiida.tools`
        to prevent the ORM namespace from getting too cluttered.

.. note:: Modules in this sub package may require the database environment to be loaded

"""

# AUTO-GENERATED

# fmt: off

from .calculations import *
from .data import *
from .dumping import *
from .graph import *
from .groups import *
from .visualization import *

__all__ = (
    'CalculationTools',
    'DELETE_LOGGER',
    'Graph',
    'GroupNotFoundError',
    'GroupNotUniqueError',
    'GroupPath',
    'InvalidPath',
    'NoGroupsInPathError',
    'Orbital',
    'RealhydrogenOrbital',
    'default_link_styles',
    'default_node_styles',
    'default_node_sublabels',
    'delete_group_nodes',
    'delete_nodes',
    'get_explicit_kpoints_path',
    'get_kpoints_path',
    'pstate_node_styles',
    'spglib_tuple_to_structure',
    'structure_to_spglib_tuple',
)

# fmt: on



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of known configuration options and methods to parse and get option values."""

from typing import Any, Dict, List, Tuple

from aiida.common.exceptions import ConfigurationError

__all__ = ('get_option', 'get_option_names', 'parse_option', 'Option')


class Option:
    """Represent a configuration option schema."""

    def __init__(self, name: str, schema: Dict[str, Any], field):
        self._name = name
        self._schema = schema
        self._field = field

    def __str__(self) -> str:
        return f'Option(name={self._name})'

    @property
    def name(self) -> str:
        return self._name

    @property
    def valid_type(self) -> Any:
        return self._field.annotation

    @property
    def schema(self) -> Dict[str, Any]:
        return self._schema

    @property
    def default(self) -> Any:
        return self._field.default

    @property
    def description(self) -> str:
        return self._field.description

    @property
    def global_only(self) -> bool:
        from .config import ProfileOptionsSchema

        return self._name.replace('.', '__') not in ProfileOptionsSchema.model_fields

    def validate(self, value: Any) -> Any:
        """Validate a value

        :param value: The input value
        :return: The output value
        :raise: ConfigurationError
        """
        from pydantic import ValidationError

        from .config import GlobalOptionsSchema

        attribute = self.name.replace('.', '__')

        try:
            # There is no straightforward way to validate a single field of a model in pydantic v2.0. The following
            # approach is the current work around, see: https://github.com/pydantic/pydantic/discussions/7367
            result = GlobalOptionsSchema.__pydantic_validator__.validate_assignment(
                GlobalOptionsSchema.model_construct(), attribute, value
            )
        except ValidationError as exception:
            messages = []
            for error in exception.errors():
                try:
                    messages.append(str(error['ctx']['error']))
                except KeyError:
                    messages.append(f"Invalid value for `{error['loc'][0]}`: {error['msg']}")

            raise ConfigurationError('\n'.join(messages)) from exception

        # Return the value from the constructed model as this will have casted the value to the right type
        return getattr(result, attribute)


def get_option_names() -> List[str]:
    """Return a list of available option names."""
    from .config import GlobalOptionsSchema

    return [key.replace('__', '.') for key in GlobalOptionsSchema.model_fields]


def get_option(name: str) -> Option:
    """Return option."""
    from .config import GlobalOptionsSchema

    options = GlobalOptionsSchema.model_fields
    option_name = name.replace('.', '__')
    if option_name not in options:
        raise ConfigurationError(f'the option {name} does not exist')
    return Option(name, GlobalOptionsSchema.model_json_schema()['properties'][option_name], options[option_name])


def parse_option(option_name: str, option_value: Any) -> Tuple[Option, Any]:
    """Parse and validate a value for a configuration option.

    :param option_name: the name of the configuration option
    :param option_value: the option value
    :return: a tuple of the option and the parsed value

    """
    option = get_option(option_name)
    value = option.validate(option_value)

    return option, value



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Data plugin representing an executable code on a remote computer.

This plugin should be used if an executable is pre-installed on a computer. The ``InstalledCode`` represents the code by
storing the absolute filepath of the relevant executable and the computer on which it is installed. The computer is
represented by an instance of :class:`aiida.orm.computers.Computer`. Each time a :class:`aiida.engine.CalcJob` is run
using an ``InstalledCode``, it will run its executable on the associated computer.
"""

from __future__ import annotations

import pathlib
import typing as t

from pydantic import ConfigDict, field_validator, model_validator

from aiida.common import exceptions
from aiida.common.lang import type_check
from aiida.common.log import override_log_level
from aiida.common.pydantic import MetadataField
from aiida.orm import Computer
from aiida.orm.entities import from_backend_entity

from .abstract import AbstractCode
from .legacy import Code

__all__ = ('InstalledCode',)


class InstalledCode(Code):
    """Data plugin representing an executable code on a remote computer."""

    _EMIT_CODE_DEPRECATION_WARNING: bool = False
    _KEY_ATTRIBUTE_FILEPATH_EXECUTABLE: str = 'filepath_executable'

    class Model(AbstractCode.Model):
        """Model describing required information to create an instance."""

        model_config = ConfigDict(arbitrary_types_allowed=True)

        computer: t.Union[str, Computer] = MetadataField(
            ...,
            title='Computer',
            description='The remote computer on which the executable resides.',
            short_name='-Y',
            priority=2,
        )
        filepath_executable: str = MetadataField(
            ...,
            title='Filepath executable',
            description='Filepath of the executable on the remote computer.',
            short_name='-X',
            priority=1,
        )

        @field_validator('label')
        @classmethod
        def validate_label_uniqueness(cls, value: str) -> str:
            """Override the validator for the ``label`` of the base class since uniqueness is defined on full label."""
            return value

        @field_validator('computer')
        @classmethod
        def validate_computer(cls, value: str) -> Computer:
            """Override the validator for the ``label`` of the base class since uniqueness is defined on full label."""
            from aiida.orm import load_computer

            try:
                return load_computer(value)
            except exceptions.NotExistent as exception:
                raise ValueError(exception) from exception

        @model_validator(mode='after')  # type: ignore[misc]
        def validate_full_label_uniqueness(self) -> AbstractCode.Model:
            """Validate that the full label does not already exist."""
            from aiida.orm import load_code

            full_label = f'{self.label}@{self.computer.label}'  # type: ignore[union-attr]

            try:
                load_code(full_label)
            except exceptions.NotExistent:
                return self
            except exceptions.MultipleObjectsError as exception:
                raise ValueError(f'Multiple codes with the label `{full_label}` already exist.') from exception
            else:
                raise ValueError(f'A code with the label `{full_label}` already exists.')

    def __init__(self, computer: Computer, filepath_executable: str, **kwargs):
        """Construct a new instance.

        :param computer: The remote computer on which the executable is located.
        :param filepath_executable: The absolute filepath of the executable on the remote computer.
        """
        super().__init__(**kwargs)
        self.computer = computer
        self.filepath_executable = filepath_executable  # type: ignore[assignment]

    def _validate(self):
        """Validate the instance by checking that a computer has been defined.

        :raises :class:`aiida.common.exceptions.ValidationError`: If the state of the node is invalid.
        """
        super(Code, self)._validate()  # Change to ``super()._validate()`` once deprecated ``Code`` class is removed.

        if not self.computer:  # type: ignore[truthy-bool]
            raise exceptions.ValidationError('The `computer` is undefined.')

        try:
            self.filepath_executable
        except TypeError as exception:
            raise exceptions.ValidationError('The `filepath_executable` is not set.') from exception

    def validate_filepath_executable(self):
        """Validate the ``filepath_executable`` attribute.

        Checks whether the executable exists on the remote computer if a transport can be opened to it. This method
        is intentionally not called in ``_validate`` as to allow the creation of ``Code`` instances whose computers can
        not yet be connected to and as to not require the overhead of opening transports in storing a new code.

        .. note:: If the ``filepath_executable`` is not an absolute path, the check is skipped.

        :raises `~aiida.common.exceptions.ValidationError`: if no transport could be opened or if the defined executable
            does not exist on the remote computer.
        """
        if not self.filepath_executable.is_absolute():
            return

        try:
            with override_log_level():  # Temporarily suppress noisy logging
                with self.computer.get_transport() as transport:
                    file_exists = transport.isfile(str(self.filepath_executable))
        except Exception as exception:
            raise exceptions.ValidationError(
                'Could not connect to the configured computer to determine whether the specified executable exists.'
            ) from exception

        if not file_exists:
            raise exceptions.ValidationError(
                f'The provided remote absolute path `{self.filepath_executable}` does not exist on the computer.'
            )

    def can_run_on_computer(self, computer: Computer) -> bool:
        """Return whether the code can run on a given computer.

        :param computer: The computer.
        :return: ``True`` if the provided computer is the same as the one configured for this code.
        """
        type_check(computer, Computer)
        return computer.pk == self.computer.pk

    def get_executable(self) -> pathlib.PurePosixPath:
        """Return the executable that the submission script should execute to run the code.

        :return: The executable to be called in the submission script.
        """
        return self.filepath_executable

    @property  # type: ignore[override]
    def computer(self) -> Computer:
        """Return the computer of this code."""
        assert self.backend_entity.computer is not None
        return from_backend_entity(Computer, self.backend_entity.computer)

    @computer.setter
    def computer(self, computer: Computer) -> None:
        """Set the computer of this code.

        :param computer: A `Computer`.
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('cannot set the computer on a stored node')

        type_check(computer, Computer, allow_none=False)
        self.backend_entity.computer = computer.backend_entity

    @property
    def full_label(self) -> str:
        """Return the full label of this code.

        The full label can be just the label itself but it can be something else. However, it at the very least has to
        include the label of the code.

        :return: The full label of the code.
        """
        return f'{self.label}@{self.computer.label}'

    @property
    def filepath_executable(self) -> pathlib.PurePosixPath:
        """Return the absolute filepath of the executable that this code represents.

        :return: The absolute filepath of the executable.
        """
        return pathlib.PurePosixPath(self.base.attributes.get(self._KEY_ATTRIBUTE_FILEPATH_EXECUTABLE))

    @filepath_executable.setter
    def filepath_executable(self, value: str) -> None:
        """Set the absolute filepath of the executable that this code represents.

        :param value: The absolute filepath of the executable.
        """
        type_check(value, str)
        self.base.attributes.set(self._KEY_ATTRIBUTE_FILEPATH_EXECUTABLE, value)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module to define custom click param type for multiple values"""

import click

__all__ = ('MultipleValueParamType',)


class MultipleValueParamType(click.ParamType):
    """An extension of click.ParamType that can parse multiple values for a given ParamType"""

    def __init__(self, param_type):
        """Construct a new instance."""
        super().__init__()
        self._param_type = param_type

        if hasattr(param_type, 'name'):
            self.name = f'{param_type.name}...'
        else:
            self.name = f'{param_type.__name__.upper()}...'

    def get_metavar(self, param):
        try:
            return self._param_type.get_metavar(param)
        except AttributeError:
            return self.name

    def convert(self, value, param, ctx):
        try:
            return tuple(self._param_type(entry) for entry in value)
        except ValueError:
            self.fail(f'could not convert {value} into type {self._param_type}')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module to define the (physical) constants used throughout the code."""

# This is the precision with which AiiDA internally will store float numbers
# In particular, before storing a number (in attributes/extras),
# when going via clean_value, AiiDA will first serialize to a string with
# this precision, and then bring it back to float. This is important to
# ensure consistency between the precision of different backends, and also
# for hashing - the hashing MUST use the same precision, so that a hash computed
# after clean_value but before storing to the database will be the same as the
# hash recomputed (e.g. via `verdi rehash`) after storing and reloading the node
# from the database.
# See also discussion on GitHub issue #2631
#
# IMPORTANT! Changing this value requires to rehash the whole database!
# So if you plan to change this, think carefully, and you might need to add
# a migration.
AIIDA_FLOAT_PRECISION = 14

# Element table, from NIST (http://www.nist.gov/pml/data/index.cfm)
# Retrieved in October 2014 for atomic numbers 1-103, and in May 2016 or atomic numbers 104-112, 114 and 116.
# In addition, element X is added to support unknown elements.
elements = {
    0: {'mass': 1.00000, 'name': 'Unknown', 'symbol': 'X'},
    1: {'mass': 1.00794, 'name': 'Hydrogen', 'symbol': 'H'},
    2: {'mass': 4.002602, 'name': 'Helium', 'symbol': 'He'},
    3: {'mass': 6.941, 'name': 'Lithium', 'symbol': 'Li'},
    4: {'mass': 9.012182, 'name': 'Beryllium', 'symbol': 'Be'},
    5: {'mass': 10.811, 'name': 'Boron', 'symbol': 'B'},
    6: {'mass': 12.0107, 'name': 'Carbon', 'symbol': 'C'},
    7: {'mass': 14.0067, 'name': 'Nitrogen', 'symbol': 'N'},
    8: {'mass': 15.9994, 'name': 'Oxygen', 'symbol': 'O'},
    9: {'mass': 18.9984032, 'name': 'Fluorine', 'symbol': 'F'},
    10: {'mass': 20.1797, 'name': 'Neon', 'symbol': 'Ne'},
    11: {'mass': 22.98977, 'name': 'Sodium', 'symbol': 'Na'},
    12: {'mass': 24.305, 'name': 'Magnesium', 'symbol': 'Mg'},
    13: {'mass': 26.981538, 'name': 'Aluminium', 'symbol': 'Al'},
    14: {'mass': 28.0855, 'name': 'Silicon', 'symbol': 'Si'},
    15: {'mass': 30.973761, 'name': 'Phosphorus', 'symbol': 'P'},
    16: {'mass': 32.065, 'name': 'Sulfur', 'symbol': 'S'},
    17: {'mass': 35.453, 'name': 'Chlorine', 'symbol': 'Cl'},
    18: {'mass': 39.948, 'name': 'Argon', 'symbol': 'Ar'},
    19: {'mass': 39.0983, 'name': 'Potassium', 'symbol': 'K'},
    20: {'mass': 40.078, 'name': 'Calcium', 'symbol': 'Ca'},
    21: {'mass': 44.955912, 'name': 'Scandium', 'symbol': 'Sc'},
    22: {'mass': 47.867, 'name': 'Titanium', 'symbol': 'Ti'},
    23: {'mass': 50.9415, 'name': 'Vanadium', 'symbol': 'V'},
    24: {'mass': 51.9961, 'name': 'Chromium', 'symbol': 'Cr'},
    25: {'mass': 54.938045, 'name': 'Manganese', 'symbol': 'Mn'},
    26: {'mass': 55.845, 'name': 'Iron', 'symbol': 'Fe'},
    27: {'mass': 58.933195, 'name': 'Cobalt', 'symbol': 'Co'},
    28: {'mass': 58.6934, 'name': 'Nickel', 'symbol': 'Ni'},
    29: {'mass': 63.546, 'name': 'Copper', 'symbol': 'Cu'},
    30: {'mass': 65.38, 'name': 'Zinc', 'symbol': 'Zn'},
    31: {'mass': 69.723, 'name': 'Gallium', 'symbol': 'Ga'},
    32: {'mass': 72.64, 'name': 'Germanium', 'symbol': 'Ge'},
    33: {'mass': 74.9216, 'name': 'Arsenic', 'symbol': 'As'},
    34: {'mass': 78.96, 'name': 'Selenium', 'symbol': 'Se'},
    35: {'mass': 79.904, 'name': 'Bromine', 'symbol': 'Br'},
    36: {'mass': 83.798, 'name': 'Krypton', 'symbol': 'Kr'},
    37: {'mass': 85.4678, 'name': 'Rubidium', 'symbol': 'Rb'},
    38: {'mass': 87.62, 'name': 'Strontium', 'symbol': 'Sr'},
    39: {'mass': 88.90585, 'name': 'Yttrium', 'symbol': 'Y'},
    40: {'mass': 91.224, 'name': 'Zirconium', 'symbol': 'Zr'},
    41: {'mass': 92.90638, 'name': 'Niobium', 'symbol': 'Nb'},
    42: {'mass': 95.96, 'name': 'Molybdenum', 'symbol': 'Mo'},
    43: {'mass': 98.0, 'name': 'Technetium', 'symbol': 'Tc'},
    44: {'mass': 101.07, 'name': 'Ruthenium', 'symbol': 'Ru'},
    45: {'mass': 102.9055, 'name': 'Rhodium', 'symbol': 'Rh'},
    46: {'mass': 106.42, 'name': 'Palladium', 'symbol': 'Pd'},
    47: {'mass': 107.8682, 'name': 'Silver', 'symbol': 'Ag'},
    48: {'mass': 112.411, 'name': 'Cadmium', 'symbol': 'Cd'},
    49: {'mass': 114.818, 'name': 'Indium', 'symbol': 'In'},
    50: {'mass': 118.71, 'name': 'Tin', 'symbol': 'Sn'},
    51: {'mass': 121.76, 'name': 'Antimony', 'symbol': 'Sb'},
    52: {'mass': 127.6, 'name': 'Tellurium', 'symbol': 'Te'},
    53: {'mass': 126.90447, 'name': 'Iodine', 'symbol': 'I'},
    54: {'mass': 131.293, 'name': 'Xenon', 'symbol': 'Xe'},
    55: {'mass': 132.9054519, 'name': 'Caesium', 'symbol': 'Cs'},
    56: {'mass': 137.327, 'name': 'Barium', 'symbol': 'Ba'},
    57: {'mass': 138.90547, 'name': 'Lanthanum', 'symbol': 'La'},
    58: {'mass': 140.116, 'name': 'Cerium', 'symbol': 'Ce'},
    59: {'mass': 140.90765, 'name': 'Praseodymium', 'symbol': 'Pr'},
    60: {'mass': 144.242, 'name': 'Neodymium', 'symbol': 'Nd'},
    61: {'mass': 145.0, 'name': 'Promethium', 'symbol': 'Pm'},
    62: {'mass': 150.36, 'name': 'Samarium', 'symbol': 'Sm'},
    63: {'mass': 151.964, 'name': 'Europium', 'symbol': 'Eu'},
    64: {'mass': 157.25, 'name': 'Gadolinium', 'symbol': 'Gd'},
    65: {'mass': 158.92535, 'name': 'Terbium', 'symbol': 'Tb'},
    66: {'mass': 162.5, 'name': 'Dysprosium', 'symbol': 'Dy'},
    67: {'mass': 164.93032, 'name': 'Holmium', 'symbol': 'Ho'},
    68: {'mass': 167.259, 'name': 'Erbium', 'symbol': 'Er'},
    69: {'mass': 168.93421, 'name': 'Thulium', 'symbol': 'Tm'},
    70: {'mass': 173.054, 'name': 'Ytterbium', 'symbol': 'Yb'},
    71: {'mass': 174.9668, 'name': 'Lutetium', 'symbol': 'Lu'},
    72: {'mass': 178.49, 'name': 'Hafnium', 'symbol': 'Hf'},
    73: {'mass': 180.94788, 'name': 'Tantalum', 'symbol': 'Ta'},
    74: {'mass': 183.84, 'name': 'Tungsten', 'symbol': 'W'},
    75: {'mass': 186.207, 'name': 'Rhenium', 'symbol': 'Re'},
    76: {'mass': 190.23, 'name': 'Osmium', 'symbol': 'Os'},
    77: {'mass': 192.217, 'name': 'Iridium', 'symbol': 'Ir'},
    78: {'mass': 195.084, 'name': 'Platinum', 'symbol': 'Pt'},
    79: {'mass': 196.966569, 'name': 'Gold', 'symbol': 'Au'},
    80: {'mass': 200.59, 'name': 'Mercury', 'symbol': 'Hg'},
    81: {'mass': 204.3833, 'name': 'Thallium', 'symbol': 'Tl'},
    82: {'mass': 207.2, 'name': 'Lead', 'symbol': 'Pb'},
    83: {'mass': 208.9804, 'name': 'Bismuth', 'symbol': 'Bi'},
    84: {'mass': 209.0, 'name': 'Polonium', 'symbol': 'Po'},
    85: {'mass': 210.0, 'name': 'Astatine', 'symbol': 'At'},
    86: {'mass': 222.0, 'name': 'Radon', 'symbol': 'Rn'},
    87: {'mass': 223.0, 'name': 'Francium', 'symbol': 'Fr'},
    88: {'mass': 226.0, 'name': 'Radium', 'symbol': 'Ra'},
    89: {'mass': 227.0, 'name': 'Actinium', 'symbol': 'Ac'},
    90: {'mass': 232.03806, 'name': 'Thorium', 'symbol': 'Th'},
    91: {'mass': 231.03588, 'name': 'Protactinium', 'symbol': 'Pa'},
    92: {'mass': 238.02891, 'name': 'Uranium', 'symbol': 'U'},
    93: {'mass': 237.0, 'name': 'Neptunium', 'symbol': 'Np'},
    94: {'mass': 244.0, 'name': 'Plutonium', 'symbol': 'Pu'},
    95: {'mass': 243.0, 'name': 'Americium', 'symbol': 'Am'},
    96: {'mass': 247.0, 'name': 'Curium', 'symbol': 'Cm'},
    97: {'mass': 247.0, 'name': 'Berkelium', 'symbol': 'Bk'},
    98: {'mass': 251.0, 'name': 'Californium', 'symbol': 'Cf'},
    99: {'mass': 252.0, 'name': 'Einsteinium', 'symbol': 'Es'},
    100: {'mass': 257.0, 'name': 'Fermium', 'symbol': 'Fm'},
    101: {'mass': 258.0, 'name': 'Mendelevium', 'symbol': 'Md'},
    102: {'mass': 259.0, 'name': 'Nobelium', 'symbol': 'No'},
    103: {'mass': 262.0, 'name': 'Lawrencium', 'symbol': 'Lr'},
    104: {'mass': 267.0, 'name': 'Rutherfordium', 'symbol': 'Rf'},
    105: {'mass': 268.0, 'name': 'Dubnium', 'symbol': 'Db'},
    106: {'mass': 271.0, 'name': 'Seaborgium', 'symbol': 'Sg'},
    107: {'mass': 272.0, 'name': 'Bohrium', 'symbol': 'Bh'},
    108: {'mass': 270.0, 'name': 'Hassium', 'symbol': 'Hs'},
    109: {'mass': 276.0, 'name': 'Meitnerium', 'symbol': 'Mt'},
    110: {'mass': 281.0, 'name': 'Darmstadtium', 'symbol': 'Ds'},
    111: {'mass': 280.0, 'name': 'Roentgenium', 'symbol': 'Rg'},
    112: {'mass': 285.0, 'name': 'Copernicium', 'symbol': 'Cn'},
    114: {'mass': 289.0, 'name': 'Flerovium', 'symbol': 'Fl'},
    116: {'mass': 293.0, 'name': 'Livermorium', 'symbol': 'Lv'},
}



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration from v0.10 to v0.11, used by ``verdi archive migrate`` command.

This migration applies the name change of the ``Computer`` attribute ``name`` to ``label``.
"""

from ..utils import update_metadata, verify_metadata_version


def migrate_v10_to_v11(metadata: dict, data: dict) -> None:
    """Migration of export files from v0.10 to v0.11."""
    old_version = '0.10'
    new_version = '0.11'

    verify_metadata_version(metadata, old_version)
    update_metadata(metadata, new_version)

    # Apply migrations
    for attributes in data.get('export_data', {}).get('Computer', {}).values():
        attributes['label'] = attributes.pop('name')

    try:
        metadata['all_fields_info']['Computer']['label'] = metadata['all_fields_info']['Computer'].pop('name')
    except KeyError:
        pass



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# ruff: noqa: N802
"""Definition of factories to load classes from the various plugin groups."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Literal, NoReturn, Tuple, Type, Union, overload

from aiida.common.exceptions import InvalidEntryPointTypeError

__all__ = (
    'BaseFactory',
    'BrokerFactory',
    'CalculationFactory',
    'CalcJobImporterFactory',
    'DataFactory',
    'DbImporterFactory',
    'GroupFactory',
    'OrbitalFactory',
    'ParserFactory',
    'SchedulerFactory',
    'StorageFactory',
    'TransportFactory',
    'WorkflowFactory',
)

if TYPE_CHECKING:
    from importlib_metadata import EntryPoint

    from aiida.brokers import Broker
    from aiida.engine import CalcJob, CalcJobImporter, WorkChain
    from aiida.orm import Data, Group
    from aiida.orm.implementation import StorageBackend
    from aiida.parsers import Parser
    from aiida.schedulers import Scheduler
    from aiida.tools.data.orbital import Orbital
    from aiida.tools.dbimporters import DbImporter
    from aiida.transports import Transport


def raise_invalid_type_error(entry_point_name: str, entry_point_group: str, valid_classes: Tuple[Any, ...]) -> NoReturn:
    """Raise an `InvalidEntryPointTypeError` with formatted message.

    :param entry_point_name: name of the entry point
    :param entry_point_group: name of the entry point group
    :param valid_classes: tuple of valid classes for the given entry point group
    :raises aiida.common.InvalidEntryPointTypeError: always
    """
    template = 'entry point `{}` registered in group `{}` is invalid because its type is not one of: {}'
    args = (entry_point_name, entry_point_group, ', '.join([e.__name__ for e in valid_classes]))
    raise InvalidEntryPointTypeError(template.format(*args))


def BaseFactory(group: str, name: str, load: bool = True) -> Union[EntryPoint, Any]:
    """Return the plugin class registered under a given entry point group and name.

    :param group: entry point group
    :param name: entry point name
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: the plugin class
    :raises aiida.common.MissingEntryPointError: entry point was not registered
    :raises aiida.common.MultipleEntryPointError: entry point could not be uniquely resolved
    :raises aiida.common.LoadingEntryPointError: entry point could not be loaded
    """
    from .entry_point import get_entry_point, load_entry_point

    if load is True:
        return load_entry_point(group, name)

    return get_entry_point(group, name)


@overload
def BrokerFactory(entry_point_name: str, load: Literal[True] = True) -> Type['Broker']: ...


@overload
def BrokerFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def BrokerFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['Broker'], Callable]:
    """Return the `Broker` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.brokers.broker.Broker`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.brokers import Broker

    entry_point_group = 'aiida.brokers'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Broker,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Broker):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def CalculationFactory(entry_point_name: str, load: Literal[True] = True) -> Union[Type['CalcJob'], Callable]: ...


@overload
def CalculationFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def CalculationFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['CalcJob'], Callable]:
    """Return the `CalcJob` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.engine.processes.calcjobs.calcjob.CalcJob`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.engine import CalcJob, calcfunction, is_process_function
    from aiida.orm import CalcFunctionNode

    entry_point_group = 'aiida.calculations'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (CalcJob, calcfunction)

    if not load:
        return entry_point

    if (isclass(entry_point) and issubclass(entry_point, CalcJob)) or (
        is_process_function(entry_point) and entry_point.node_class is CalcFunctionNode  # type: ignore[union-attr]
    ):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def CalcJobImporterFactory(entry_point_name: str, load: Literal[True] = True) -> Type['CalcJobImporter']: ...


@overload
def CalcJobImporterFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def CalcJobImporterFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['CalcJobImporter']]:
    """Return the plugin registered under the given entry point.

    :param entry_point_name: the entry point name.
    :return: the loaded :class:`~aiida.engine.processes.calcjobs.importer.CalcJobImporter` plugin.
    :raises ``aiida.common.InvalidEntryPointTypeError``: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.engine import CalcJobImporter

    entry_point_group = 'aiida.calculations.importers'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (CalcJobImporter,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, CalcJobImporter):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def DataFactory(entry_point_name: str, load: Literal[True] = True) -> Type['Data']: ...


@overload
def DataFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def DataFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['Data']]:
    """Return the `Data` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.orm.nodes.data.data.Data`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.orm import Data

    entry_point_group = 'aiida.data'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Data,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Data):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def DbImporterFactory(entry_point_name: str, load: Literal[True] = True) -> Type['DbImporter']: ...


@overload
def DbImporterFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def DbImporterFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['DbImporter']]:
    """Return the `DbImporter` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.tools.dbimporters.baseclasses.DbImporter`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.tools.dbimporters import DbImporter

    entry_point_group = 'aiida.tools.dbimporters'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (DbImporter,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, DbImporter):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def GroupFactory(entry_point_name: str, load: Literal[True] = True) -> Type['Group']: ...


@overload
def GroupFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def GroupFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['Group']]:
    """Return the `Group` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.orm.groups.Group`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.orm import Group

    entry_point_group = 'aiida.groups'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Group,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Group):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def OrbitalFactory(entry_point_name: str, load: Literal[True] = True) -> Type['Orbital']: ...


@overload
def OrbitalFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def OrbitalFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['Orbital']]:
    """Return the `Orbital` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.tools.data.orbital.orbital.Orbital`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.tools.data.orbital import Orbital

    entry_point_group = 'aiida.tools.data.orbitals'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Orbital,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Orbital):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def ParserFactory(entry_point_name: str, load: Literal[True] = True) -> Type['Parser']: ...


@overload
def ParserFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def ParserFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['Parser']]:
    """Return the `Parser` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.parsers.parser.Parser`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.parsers import Parser

    entry_point_group = 'aiida.parsers'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Parser,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Parser):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def SchedulerFactory(entry_point_name: str, load: Literal[True] = True) -> Type['Scheduler']: ...


@overload
def SchedulerFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def SchedulerFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['Scheduler']]:
    """Return the `Scheduler` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.schedulers.scheduler.Scheduler`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.schedulers import Scheduler

    entry_point_group = 'aiida.schedulers'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Scheduler,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Scheduler):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def StorageFactory(entry_point_name: str, load: Literal[True] = True) -> Type['StorageBackend']: ...


@overload
def StorageFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def StorageFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['StorageBackend']]:
    """Return the ``StorageBackend`` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.orm.implementation.storage_backend.StorageBackend`.
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.orm.implementation import StorageBackend

    entry_point_group = 'aiida.storage'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (StorageBackend,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, StorageBackend):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def TransportFactory(entry_point_name: str, load: Literal[True] = True) -> Type['Transport']: ...


@overload
def TransportFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def TransportFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['Transport']]:
    """Return the `Transport` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.transports import Transport

    entry_point_group = 'aiida.transports'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Transport,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Transport):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def WorkflowFactory(entry_point_name: str, load: Literal[True] = True) -> Union[Type['WorkChain'], Callable]: ...


@overload
def WorkflowFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint: ...


def WorkflowFactory(entry_point_name: str, load: bool = True) -> Union[EntryPoint, Type['WorkChain'], Callable]:
    """Return the `WorkChain` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.engine.processes.workchains.workchain.WorkChain` or a `workfunction`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from inspect import isclass

    from aiida.engine import WorkChain, is_process_function, workfunction
    from aiida.orm import WorkFunctionNode

    entry_point_group = 'aiida.workflows'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (WorkChain, workfunction)

    if not load:
        return entry_point

    if (isclass(entry_point) and issubclass(entry_point, WorkChain)) or (
        is_process_function(entry_point) and entry_point.node_class is WorkFunctionNode  # type: ignore[union-attr]
    ):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tool for handling data."""

# AUTO-GENERATED

# fmt: off

from .array import *
from .orbital import *
from .structure import *

__all__ = (
    'Orbital',
    'RealhydrogenOrbital',
    'get_explicit_kpoints_path',
    'get_kpoints_path',
    'spglib_tuple_to_structure',
    'structure_to_spglib_tuple',
)

# fmt: on



#!/usr/bin/env python
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Validate consistency of versions and dependencies.

Validates consistency of

 * pyproject.toml
 * environment.yml

"""

import os

import click

SCRIPT_PATH = os.path.split(os.path.realpath(__file__))[0]
ROOT_DIR = os.path.join(SCRIPT_PATH, os.pardir)


def determine_block_positions(lines, block_start_marker, block_end_marker):
    """Determine the line indices of a block in a list of lines indicated by a given start and end marker.

    :param lines: list of strings
    :param block_start_marker: string marking the beginning of the block
    :param block_end_marker: string marking the end of the block
    :return: tuple of two integers representing the line indices that indicate the start and end of the block
    """
    block_start_index = -1
    block_end_index = -1

    for line_number, line in enumerate(lines):
        if block_start_marker in line:
            block_start_index = line_number + 1

        if block_end_marker in line:
            block_end_index = line_number
            break

    if block_start_index < 0 or block_end_index < 0:
        raise RuntimeError('failed to determine the starting or end point of the block')

    return block_start_index, block_end_index


def replace_line_block(lines, block, index_start, index_end):
    """Replace a block of lines between two line indices with a new set of lines.

    :param lines: list of lines representing the whole file
    :param block: list of lines representing the new block that should be inserted after old block is removed
    :param index_start: start of the block to be removed
    :param index_end: end of the block to be removed
    :return: list of lines with block of lines replaced
    """
    # Slice out the old block by removing the lines between the markers of the block
    lines = lines[:index_start] + lines[index_end:]

    # Now insert the new block starting at the beginning of the original block
    lines[index_start:index_start] = block

    return lines


def replace_block_in_file(filepath, block_start_marker, block_end_marker, block):
    """Replace a block of text between the given string markers with the provided new block of lines.

    :param filepath: absolute path of the file
    :param block_start_marker: string marking the beginning of the block
    :param block_end_marker: string marking the end of the block
    :param block: list of lines representing the new block that should be inserted after old block is removed
    """
    with open(filepath, encoding='utf8') as handle:
        lines = handle.readlines()

    try:
        index_start, index_end = determine_block_positions(lines, block_start_marker, block_end_marker)
    except RuntimeError as exception:
        raise RuntimeError(f'problem rewriting file `{filepath}`:: {exception}')

    lines = replace_line_block(lines, block, index_start, index_end)

    with open(filepath, 'w', encoding='utf8') as handle:
        for line in lines:
            handle.write(line)


@click.group()
def cli():
    pass


@cli.command('verdi-autodocs')
def validate_verdi_documentation():
    """Auto-generate the documentation for `verdi` through `click`."""
    from aiida.cmdline.commands.cmd_verdi import verdi
    from click import Context

    width = 90  # The maximum width of the formatted help strings in characters

    # Set the `verdi data` command to isolated mode such that external plugin commands are not discovered
    ctx = Context(verdi, terminal_width=width)
    command = verdi.get_command(ctx, 'data')
    command.set_exclude_external_plugins(True)

    # Replacing the block with the commands of `verdi`
    filepath_verdi_commands = os.path.join(ROOT_DIR, 'docs', 'source', 'reference', 'command_line.rst')
    commands_block_start_marker = '.. _reference:command-line:verdi:'
    commands_block_end_marker = '.. END_OF_VERDI_COMMANDS_MARKER'

    # Generate the new block with the command help strings
    header = 'Commands'
    message = 'Below is a list with all available subcommands.'
    block = [f"{header}\n{'=' * len(header)}\n{message}\n\n"]

    for name, command in sorted(verdi.commands.items()):
        if name == 'tui':
            # This command is only generated when the optional dependency ``trogon`` is installed. It provides a TUI
            # version of ``verdi``. However, since it is optional, if a development environment does not have it
            # installed, this check will always fail as the generated docs are different. Since ``trogon`` significantly
            # slows down tab-completion of ``verdi``, many dev environments do not want to have it installed. As a
            # workaround, we are excluding this command from the automatically generated reference documentation.
            continue

        ctx = click.Context(command, terminal_width=width)

        header_label = f'.. _reference:command-line:verdi-{name}:'
        header_string = f'``verdi {name}``'
        header_underline = '-' * len(header_string)

        block.append(f'{header_label}\n\n')
        block.append(f'{header_string}\n')
        block.append(f'{header_underline}\n\n')
        block.append('.. code:: console\n\n')  # Mark the beginning of a literal block
        for line in ctx.get_help().split('\n'):
            if line:
                block.append(f'    {line}\n')
            else:
                block.append('\n')
        block.append('\n\n')

    # New block should start and end with an empty line after and before the literal block marker
    block.insert(0, '\n')
    block.append('\n')

    replace_block_in_file(filepath_verdi_commands, commands_block_start_marker, commands_block_end_marker, block)


if __name__ == '__main__':
    cli()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Base implementation of `WorkChain` class that implements a simple automated restart mechanism for sub processes."""

import functools
from inspect import getmembers
from types import FunctionType
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Tuple, Type, Union

from aiida import orm
from aiida.common import AttributeDict
from aiida.common.links import LinkType
from aiida.common.warnings import warn_deprecation

from .context import ToContext, append_
from .utils import ProcessHandlerReport, process_handler
from .workchain import WorkChain

if TYPE_CHECKING:
    from aiida.engine.processes import ExitCode, PortNamespace, Process, ProcessSpec

__all__ = ('BaseRestartWorkChain',)


def validate_handler_overrides(
    process_class: 'BaseRestartWorkChain', handler_overrides: Optional[orm.Dict], ctx: 'PortNamespace'
) -> Optional[str]:
    """Validator for the ``handler_overrides`` input port of the ``BaseRestartWorkChain``.

    The ``handler_overrides`` should be a dictionary where keys are strings that are the name of a process handler, i.e.
    an instance method of the ``process_class`` that has been decorated with the ``process_handler`` decorator. The
    values should be a dictionary that can specify the keys ``enabled`` and ``priority``.

    .. note:: the normal signature of a port validator is ``(value, ctx)`` but since for the validation here we need a
        reference to the process class, we add it and the class is bound to the method in the port declaration in the
        ``define`` method.

    :param process_class: the ``BaseRestartWorkChain`` (sub) class
    :param handler_overrides: the input ``Dict`` node
    :param ctx: the ``PortNamespace`` in which the port is embedded
    """
    if not handler_overrides:
        return None

    for handler, overrides in handler_overrides.get_dict().items():
        if not isinstance(handler, str):
            return f'The key `{handler}` is not a string.'

        if not process_class.is_process_handler(handler):
            return f'The key `{handler}` is not a process handler of {process_class}'

        if not isinstance(overrides, (bool, dict)):
            return f'The value of key `{handler}` is not a boolean or dictionary.'

        if isinstance(overrides, bool):
            warn_deprecation(
                'Setting a boolean as value for `handler_overrides` is deprecated. Use '
                "`{'handler_name': {'enabled': " + f'{overrides}' + '}` instead.',
                version=3,
            )

        if isinstance(overrides, dict):
            for key in overrides.keys():
                if key not in ['enabled', 'priority']:
                    return f'The value of key `{handler}` contain keys `{key}` which is not supported.'

    return None


class BaseRestartWorkChain(WorkChain):
    """Base restart work chain.

    This work chain serves as the starting point for more complex work chains that will be designed to run a sub process
    that might need multiple restarts to come to a successful end. These restarts may be necessary because a single
    process run is not sufficient to achieve a fully converged result, or certain errors maybe encountered which
    are recoverable.

    This work chain implements the most basic functionality to achieve this goal. It will launch the sub process,
    restarting until it is completed successfully or the maximum number of iterations is reached. After completion of
    the sub process it will be inspected, and a list of process handlers are called successively. These process handlers
    are defined as class methods that are decorated with :meth:`~aiida.engine.process_handler`.

    The idea is to sub class this work chain and leverage the generic error handling that is implemented in the few
    outline methods. The minimally required outline would look something like the following::

        cls.setup
        while_(cls.should_run_process)(
            cls.run_process,
            cls.inspect_process,
        )

    Each of these methods can of course be overriden but they should be general enough to fit most process cycles. The
    `run_process` method will take the inputs for the process from the context under the key `inputs`. The user should,
    therefore, make sure that before the `run_process` method is called, that the to be used inputs are stored under
    `self.ctx.inputs`. One can update the inputs based on the results from a prior process by calling an outline method
    just before the `run_process` step, for example::

        cls.setup
        while_(cls.should_run_process)(
            cls.prepare_inputs,
            cls.run_process,
            cls.inspect_process,
        )

    Where in the `prepare_calculation` method, the inputs dictionary at `self.ctx.inputs` is updated before the next
    process will be run with those inputs.

    The `_process_class` attribute should be set to the `Process` class that should be run in the loop.
    Finally, to define handlers that will be called during the `inspect_process` simply define a class method with the
    signature `(self, node)` and decorate it with the `process_handler` decorator, for example::

        @process_handler
        def handle_problem(self, node):
            if some_problem:
                self.ctx.inputs = improved_inputs
                return ProcessHandlerReport()

    The `process_handler` and `ProcessHandlerReport` support various arguments to control the flow of the logic of the
    `inspect_process`. Refer to their respective documentation for details.
    """

    _process_class: Optional[Type['Process']] = None
    _considered_handlers_extra = 'considered_handlers'

    @property
    def process_class(self) -> Type['Process']:
        """Return the process class to run in the loop."""
        from ..process import Process

        if self._process_class is None or not issubclass(self._process_class, Process):
            raise ValueError('no valid Process class defined for `_process_class` attribute')
        return self._process_class

    @classmethod
    def define(cls, spec: 'ProcessSpec') -> None:  # type: ignore[override]
        """Define the process specification."""
        super().define(spec)
        spec.input(
            'max_iterations',
            valid_type=orm.Int,
            default=lambda: orm.Int(5),
            help='Maximum number of iterations the work chain will restart the process to finish successfully.',
        )
        spec.input(
            'clean_workdir',
            valid_type=orm.Bool,
            default=lambda: orm.Bool(False),
            help='If `True`, work directories of all called calculation jobs will be cleaned at the end of execution.',
        )
        spec.input(
            'handler_overrides',
            valid_type=orm.Dict,
            required=False,
            validator=functools.partial(validate_handler_overrides, cls),
            serializer=orm.to_aiida_type,
            help='Mapping where keys are process handler names and the values are a dictionary, where each dictionary '
            'can define the ``enabled`` and ``priority`` key, which can be used to toggle the values set on '
            'the original process handler declaration.',
        )
        spec.exit_code(301, 'ERROR_SUB_PROCESS_EXCEPTED', message='The sub process excepted.')
        spec.exit_code(302, 'ERROR_SUB_PROCESS_KILLED', message='The sub process was killed.')
        spec.exit_code(
            401, 'ERROR_MAXIMUM_ITERATIONS_EXCEEDED', message='The maximum number of iterations was exceeded.'
        )
        spec.exit_code(
            402,
            'ERROR_SECOND_CONSECUTIVE_UNHANDLED_FAILURE',
            message='The process failed for an unknown reason, twice in a row.',
        )

    def setup(self) -> None:
        """Initialize context variables that are used during the logical flow of the `BaseRestartWorkChain`."""
        overrides = (
            self.inputs.handler_overrides.get_dict() if (self.inputs and 'handler_overrides' in self.inputs) else {}
        )
        self.ctx.handler_overrides = overrides
        self.ctx.process_name = self.process_class.__name__
        self.ctx.unhandled_failure = False
        self.ctx.is_finished = False
        self.ctx.iteration = 0

    def should_run_process(self) -> bool:
        """Return whether a new process should be run.

        This is the case as long as the last process has not finished successfully and the maximum number of restarts
        has not yet been exceeded.
        """
        max_iterations = self.inputs.max_iterations.value
        return not self.ctx.is_finished and self.ctx.iteration < max_iterations

    def run_process(self) -> ToContext:
        """Run the next process, taking the input dictionary from the context at `self.ctx.inputs`."""
        self.ctx.iteration += 1

        try:
            unwrapped_inputs = self.ctx.inputs
        except AttributeError:
            raise AttributeError('no process input dictionary was defined in `self.ctx.inputs`')

        # Set the `CALL` link label
        unwrapped_inputs.setdefault('metadata', {})['call_link_label'] = f'iteration_{self.ctx.iteration:02d}'

        inputs = self._wrap_bare_dict_inputs(self.process_class.spec().inputs, unwrapped_inputs)
        node = self.submit(self.process_class, **inputs)

        # Add a new empty list to the `BaseRestartWorkChain._considered_handlers_extra` extra. This will contain the
        # name and return value of all class methods, decorated with `process_handler`, that are called during
        # the `inspect_process` outline step.
        considered_handlers = self.node.base.extras.get(self._considered_handlers_extra, [])
        considered_handlers.append([])
        self.node.base.extras.set(self._considered_handlers_extra, considered_handlers)

        self.report(f'launching {self.ctx.process_name}<{node.pk}> iteration #{self.ctx.iteration}')

        return ToContext(children=append_(node))

    def inspect_process(self) -> Optional['ExitCode']:
        """Analyse the results of the previous process and call the handlers when necessary.

        If the process is excepted or killed, the work chain will abort. Otherwise any attached handlers will be called
        in order of their specified priority. If the process was failed and no handler returns a report indicating that
        the error was handled, it is considered an unhandled process failure and the process is relaunched. If this
        happens twice in a row, the work chain is aborted. In the case that at least one handler returned a report the
        following matrix determines the logic that is followed:

            Process  Handler    Handler     Action
            result   report?    exit code
            -----------------------------------------
            Success      yes        == 0     Restart
            Success      yes        != 0     Abort
            Failed       yes        == 0     Restart
            Failed       yes        != 0     Abort

        If no handler returned a report and the process finished successfully, the work chain's work is considered done
        and it will move on to the next step that directly follows the `while` conditional, if there is one defined in
        the outline.
        """
        node = self.ctx.children[self.ctx.iteration - 1]

        if node.is_excepted:
            return self.exit_codes.ERROR_SUB_PROCESS_EXCEPTED

        if node.is_killed:
            return self.exit_codes.ERROR_SUB_PROCESS_KILLED

        last_report = None

        # Sort the handlers with a priority defined, based on their priority in reverse order
        for _, handler in sorted(self.get_process_handlers_by_priority(), key=lambda e: e[0], reverse=True):
            # Even though the ``handler`` is an instance method, the ``get_process_handlers_by_priority`` method returns
            # unbound methods so we have to pass in ``self`` manually. Also, always pass the ``node`` as an argument
            # because the ``process_handler`` decorator with which the handler is decorated relies on this behavior.
            report = handler(self, node)

            if report is not None and not isinstance(report, ProcessHandlerReport):
                name = handler.__name__
                raise RuntimeError(f'handler `{name}` returned a value that is not a ProcessHandlerReport')

            # If an actual report was returned, save it so it is not overridden by next handler returning `None`
            if report:
                last_report = report

            # After certain handlers, we may want to skip all other handlers
            if report and report.do_break:
                break

        report_args = (self.ctx.process_name, node.pk)

        # If the process failed and no handler returned a report we consider it an unhandled failure
        if node.is_failed and not last_report:
            if self.ctx.unhandled_failure:
                template = '{}<{}> failed and error was not handled for the second consecutive time, aborting'
                self.report(template.format(*report_args))
                return self.exit_codes.ERROR_SECOND_CONSECUTIVE_UNHANDLED_FAILURE

            self.ctx.unhandled_failure = True
            self.report('{}<{}> failed and error was not handled, restarting once more'.format(*report_args))
            return None

        # Here either the process finished successful or at least one handler returned a report so it can no longer be
        # considered to be an unhandled failed process and therefore we reset the flag
        self.ctx.unhandled_failure = False

        # If at least one handler returned a report, the action depends on its exit code and that of the process itself
        if last_report:
            if node.is_finished_ok and last_report.exit_code.status == 0:
                template = '{}<{}> finished successfully but a handler was triggered, restarting'
            elif node.is_failed and last_report.exit_code.status == 0:
                template = '{}<{}> failed but a handler dealt with the problem, restarting'
            elif node.is_finished_ok and last_report.exit_code.status != 0:
                template = '{}<{}> finished successfully but a handler detected an unrecoverable problem, aborting'
            elif node.is_failed and last_report.exit_code.status != 0:
                template = '{}<{}> failed but a handler detected an unrecoverable problem, aborting'

            self.report(template.format(*report_args))

            return last_report.exit_code

        # Otherwise the process was successful and no handler returned anything so we consider the work done
        self.ctx.is_finished = True

        return None

    def get_outputs(self, node) -> Mapping[str, orm.Node]:
        """Return a mapping of the outputs that should be attached as outputs to the work chain.

        By default this method returns the outputs of the last completed calculation job. This method can be overridden
        if the implementation wants to update those outputs before attaching them. Make sure that if the content of an
        output node is modified that this is done through a calcfunction in order to not lose the provenance.
        """
        return self.exposed_outputs(node, self.process_class)

    def results(self) -> Optional['ExitCode']:
        """Attach the outputs specified in the output specification from the last completed process."""
        node = self.ctx.children[self.ctx.iteration - 1]

        # We check the `is_finished` attribute of the work chain and not the successfulness of the last process
        # because the error handlers in the last iteration can have qualified a "failed" process as satisfactory
        # for the outcome of the work chain and so have marked it as `is_finished=True`.
        max_iterations = self.inputs.max_iterations.value
        if not self.ctx.is_finished and self.ctx.iteration >= max_iterations:
            self.report(
                f'reached the maximum number of iterations {max_iterations}: '
                f'last ran {self.ctx.process_name}<{node.pk}>'
            )
            return self.exit_codes.ERROR_MAXIMUM_ITERATIONS_EXCEEDED

        self.report(f'work chain completed after {self.ctx.iteration} iterations')
        self._attach_outputs(node)
        return None

    def _attach_outputs(self, node) -> Mapping[str, orm.Node]:
        """Attach the outputs of the given calculation job to the work chain.

        :param node: The ``CalcJobNode`` whose outputs to attach.
        :returns: The mapping of output nodes that were attached.
        """
        outputs = self.get_outputs(node)
        existing_outputs = self.node.base.links.get_outgoing(link_type=LinkType.RETURN).all_link_labels()

        for name, port in self.spec().outputs.items():
            try:
                output = outputs[name]
            except KeyError:
                if port.required:
                    self.report(
                        f'required output `{name}` was not an output of {self.ctx.process_name}<{node.pk}> '
                        f'(or an incorrect class/output is being exposed).'
                    )
            else:
                if name in existing_outputs:
                    self.logger.info(f'output `{name}` was already attached, skipping.')
                else:
                    self.out(name, output)

        return outputs

    def __init__(self, *args, **kwargs) -> None:
        """Construct the instance."""
        super().__init__(*args, **kwargs)

        # try retrieving process class
        self.process_class

    @classmethod
    def is_process_handler(cls, process_handler_name: Union[str, FunctionType]) -> bool:
        """Return whether the given method name corresponds to a process handler of this class.

        :param process_handler_name: string name of the instance method
        :return: boolean, True if corresponds to process handler, False otherwise
        """
        if isinstance(process_handler_name, str):
            handler = getattr(cls, process_handler_name, {})
        else:
            handler = process_handler_name

        return getattr(handler, 'decorator', None) == process_handler

    @classmethod
    def get_process_handlers(cls) -> List[FunctionType]:
        return [method[1] for method in getmembers(cls) if cls.is_process_handler(method[1])]

    def get_process_handlers_by_priority(self) -> List[Tuple[int, FunctionType]]:
        """Return list of process handlers where overrides from ``inputs.handler_overrides`` are taken into account."""
        handlers = []

        for handler in self.get_process_handlers():
            overrides = self.ctx.handler_overrides.get(handler.__name__, {})

            enabled = None
            priority = None

            if isinstance(overrides, bool):
                enabled = overrides
            else:
                enabled = overrides.get('enabled')
                priority = overrides.get('priority')

            if enabled is False or (enabled is None and not handler.enabled):  # type: ignore[attr-defined]
                continue

            handlers.append((priority or handler.priority, handler))  # type: ignore[attr-defined]

        return handlers

    def on_terminated(self):
        """Clean the working directories of all child calculation jobs if `clean_workdir=True` in the inputs."""
        super().on_terminated()

        if self.inputs.clean_workdir.value is False:
            self.report('remote folders will not be cleaned')
            return

        cleaned_calcs = []

        for called_descendant in self.node.called_descendants:
            if isinstance(called_descendant, orm.CalcJobNode):
                try:
                    called_descendant.outputs.remote_folder._clean()
                    cleaned_calcs.append(str(called_descendant.pk))
                except (OSError, OSError, KeyError):
                    pass

        if cleaned_calcs:
            self.report(f"cleaned remote folders of calculations: {' '.join(cleaned_calcs)}")

    def _wrap_bare_dict_inputs(self, port_namespace: 'PortNamespace', inputs: Dict[str, Any]) -> AttributeDict:
        """Wrap bare dictionaries in `inputs` in a `Dict` node if dictated by the corresponding inputs portnamespace.

        :param port_namespace: a `PortNamespace`
        :param inputs: a dictionary of inputs intended for submission of the process
        :return: an attribute dictionary with all bare dictionaries wrapped in `Dict` if dictated by the port namespace
        """
        from aiida.engine.processes import PortNamespace

        wrapped = {}

        for key, value in inputs.items():
            if key not in port_namespace:
                wrapped[key] = value
                continue

            port = port_namespace[key]
            valid_types = port.valid_type if isinstance(port.valid_type, (list, tuple)) else (port.valid_type,)  # type: ignore[redundant-expr]

            if isinstance(port, PortNamespace):
                wrapped[key] = self._wrap_bare_dict_inputs(port, value)
            elif orm.Dict in valid_types and isinstance(value, dict):
                wrapped[key] = orm.Dict(dict=value)
            else:
                wrapped[key] = value

        return AttributeDict(wrapped)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Upper level SQLAlchemy migration funcitons."""

from alembic import context


def run_migrations_online():
    """Run migrations in 'online' mode.

    The connection should have been passed to the config, which we use to configue the migration context.
    """
    from aiida.storage.psql_dos.models.base import get_orm_metadata

    config = context.config

    connection = config.attributes.get('connection', None)
    aiida_profile = config.attributes.get('aiida_profile', None)
    on_version_apply = config.attributes.get('on_version_apply', None)

    if connection is None:
        from aiida.common.exceptions import ConfigurationError

        raise ConfigurationError('An initialized connection is expected for the AiiDA online migrations.')
    if aiida_profile is None:
        from aiida.common.exceptions import ConfigurationError

        raise ConfigurationError('An aiida_profile is expected for the AiiDA online migrations.')

    context.configure(
        connection=connection,
        target_metadata=get_orm_metadata(),
        transaction_per_migration=True,
        aiida_profile=aiida_profile,
        on_version_apply=on_version_apply,
    )

    context.run_migrations()


try:
    if context.is_offline_mode():
        raise NotImplementedError('This feature is not currently supported.')

    run_migrations_online()
except NameError:
    # This will occur in an environment that is just compiling the documentation
    pass



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Remove the `node.` prefix from `db_dbnode.type`

Final data migration for `Nodes` after `aiida.orm.nodes` reorganization was finalized to remove the `node.` prefix

Note, this is identical to the django_0028 migration.

Revision ID: 61fc0913fae9
Revises: ce56d84bcc35
Create Date: 2019-02-16 15:32:42.745450

"""

from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = '61fc0913fae9'
down_revision = 'ce56d84bcc35'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()

    # The `node.` prefix is being dropped from the node type string
    statement = text(
        r"""
        UPDATE db_dbnode
        SET type = regexp_replace(type, '^node.data.', 'data.')
        WHERE type LIKE 'node.data.%';

        UPDATE db_dbnode
        SET type = regexp_replace(type, '^node.process.', 'process.')
        WHERE type LIKE 'node.process.%';
    """
    )
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of 61fc0913fae9.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Serialisation functions for AiiDA types

WARNING: Changing the representation of things here may break people's current saved e.g. things like
checkpoints and messages in the RabbitMQ queue so do so with caution.  It is fine to add representers
for new types though.
"""

from __future__ import annotations

import inspect
from dataclasses import asdict, is_dataclass
from enum import Enum
from functools import partial
from typing import Any, Protocol, Type, overload

import yaml
from plumpy import Bundle, get_object_loader
from plumpy.utils import AttributesFrozendict

from aiida import orm
from aiida.common import AttributeDict
from aiida.orm.utils.managers import NodeLinksManager

_ENUM_TAG = '!enum'
_DATACLASS_TAG = '!dataclass'
_NODE_TAG = '!aiida_node'
_NODE_LINKS_MANAGER_TAG = '!aiida_node_links_manager'
_GROUP_TAG = '!aiida_group'
_COMPUTER_TAG = '!aiida_computer'
_ATTRIBUTE_DICT_TAG = '!aiida_attributedict'
_PLUMPY_ATTRIBUTES_FROZENDICT_TAG = '!plumpy:attributes_frozendict'
_PLUMPY_BUNDLE = '!plumpy:bundle'


def represent_enum(dumper: yaml.Dumper, enum: Enum) -> yaml.ScalarNode:
    """Represent an arbitrary enum in yaml."""
    loader = get_object_loader()
    return dumper.represent_scalar(_ENUM_TAG, f'{loader.identify_object(enum)}|{enum.value}')


def enum_constructor(loader: yaml.Loader, serialized: yaml.Node) -> Enum:
    """Construct an enum from the serialized representation."""
    deserialized: str = loader.construct_scalar(serialized)  # type: ignore[arg-type,assignment]
    identifier, value = deserialized.split('|')
    cls = get_object_loader().load_object(identifier)
    enum = cls(value)
    return enum


def represent_dataclass(dumper: yaml.Dumper, obj: Any) -> yaml.MappingNode:
    """Represent an arbitrary dataclass in yaml."""
    loader = get_object_loader()
    data = {
        '__type__': loader.identify_object(obj.__class__),
        '__fields__': asdict(obj),
    }
    return dumper.represent_mapping(_DATACLASS_TAG, data)


def dataclass_constructor(loader: yaml.Loader, serialized: yaml.Node) -> Any:
    """Construct a dataclass from the serialized representation."""
    deserialized = loader.construct_mapping(serialized, deep=True)  # type: ignore[arg-type]
    identifier = deserialized['__type__']
    cls = get_object_loader().load_object(identifier)
    data = deserialized['__fields__']
    return cls(**data)


def represent_node(dumper: yaml.Dumper, node: orm.Node) -> yaml.ScalarNode:
    """Represent a node in yaml."""
    if not node.is_stored:
        raise ValueError(f'node {type(node)}<{node.uuid}> cannot be represented because it is not stored')
    return dumper.represent_scalar(_NODE_TAG, f'{node.uuid}')


def node_constructor(loader: yaml.Loader, node: yaml.Node) -> orm.Node:
    """Load a node from the yaml representation."""
    yaml_node = loader.construct_scalar(node)  # type: ignore[arg-type]
    return orm.load_node(uuid=yaml_node)


def represent_node_links_manager(dumper: yaml.Dumper, node_links_manager: NodeLinksManager) -> yaml.MappingNode:
    """Represent a link in yaml."""
    data = {
        'incoming': node_links_manager._incoming,
        'link_type': node_links_manager._link_type.value,
        'node': node_links_manager._node.uuid,
    }
    return dumper.represent_mapping(_NODE_LINKS_MANAGER_TAG, data)


def node_links_manager_constructor(loader: yaml.Loader, node_links_manager: yaml.Node) -> NodeLinksManager:
    """Load a link from the yaml representation."""
    from aiida.common.links import LinkType

    yaml_node_links_manager = loader.construct_mapping(node_links_manager)  # type: ignore[arg-type]
    link_type = LinkType(yaml_node_links_manager['link_type'])
    node = orm.load_node(uuid=yaml_node_links_manager['node'])
    incoming = yaml_node_links_manager['incoming']
    return NodeLinksManager(node=node, link_type=link_type, incoming=incoming)


def represent_group(dumper: yaml.Dumper, group: orm.Group) -> yaml.ScalarNode:
    """Represent a group in yaml."""
    if not group.is_stored:
        raise ValueError(f'group {group} cannot be represented because it is not stored')
    return dumper.represent_scalar(_GROUP_TAG, f'{group.uuid}')


def group_constructor(loader: yaml.Loader, group: yaml.Node) -> orm.Group:
    """Load a group from the yaml representation."""
    yaml_node = loader.construct_scalar(group)  # type: ignore[arg-type]
    return orm.load_group(uuid=yaml_node)


def represent_computer(dumper: yaml.Dumper, computer: orm.Computer) -> yaml.ScalarNode:
    """Represent a computer in yaml."""
    if not computer.is_stored:
        raise ValueError(f'computer {computer} cannot be represented because it is not stored')
    return dumper.represent_scalar(_COMPUTER_TAG, f'{computer.uuid}')


def computer_constructor(loader: yaml.Loader, computer: yaml.Node) -> orm.Computer:
    """Load a computer from the yaml representation."""
    yaml_node = loader.construct_scalar(computer)  # type: ignore[arg-type]
    return orm.Computer.collection.get(uuid=yaml_node)


def represent_mapping(tag: str, dumper: yaml.Dumper, mapping: Any) -> yaml.MappingNode:
    """Represent a mapping in yaml."""
    return dumper.represent_mapping(tag, mapping)


class _MappingType(Protocol):
    def __init__(self, mapping: dict) -> None: ...


def mapping_constructor(
    mapping_type: Type[_MappingType], loader: yaml.Loader, mapping: yaml.MappingNode
) -> _MappingType:
    """Construct a mapping from the representation."""
    yaml_node = loader.construct_mapping(mapping, deep=True)
    return mapping_type(yaml_node)


def represent_bundle(dumper: yaml.Dumper, bundle: Bundle) -> yaml.MappingNode:
    """Represent an `plumpy.Bundle` in yaml."""
    as_dict = dict(bundle)
    return dumper.represent_mapping(_PLUMPY_BUNDLE, as_dict)


def bundle_constructor(loader: yaml.Loader, bundle: yaml.Node) -> Bundle:
    """Construct an `plumpy.Bundle` from the representation."""
    yaml_node = loader.construct_mapping(bundle)  # type: ignore[arg-type]
    bundle_inst = Bundle.__new__(Bundle)
    bundle_inst.update(yaml_node)
    return bundle_inst


class AiiDADumper(yaml.Dumper):
    """Custom AiiDA yaml dumper.

    Needed so that we don't have to encode each type in the AiiDA graph hierarchy separately using a custom representer.
    """

    def represent_data(self, data):
        if isinstance(data, orm.Node):
            return represent_node(self, data)
        if isinstance(data, NodeLinksManager):
            return represent_node_links_manager(self, data)
        if isinstance(data, orm.Computer):
            return represent_computer(self, data)
        if isinstance(data, orm.Group):
            return represent_group(self, data)
        if is_dataclass(data) and not inspect.isclass(data):
            return represent_dataclass(self, data)

        return super().represent_data(data)


class AiiDALoader(yaml.Loader):
    """AiiDA specific yaml loader

    .. note:: The `AiiDALoader` should only be used on trusted input, since it uses the `yaml.Loader` which is not safe.
        When importing a shared database, we strip all process node checkpoints to avoid this being a security risk.
    """


yaml.add_representer(Enum, represent_enum, Dumper=AiiDADumper)
yaml.add_representer(Bundle, represent_bundle, Dumper=AiiDADumper)
yaml.add_representer(AttributeDict, partial(represent_mapping, _ATTRIBUTE_DICT_TAG), Dumper=AiiDADumper)
yaml.add_constructor(_ATTRIBUTE_DICT_TAG, partial(mapping_constructor, AttributeDict), Loader=AiiDALoader)
yaml.add_representer(
    AttributesFrozendict, partial(represent_mapping, _PLUMPY_ATTRIBUTES_FROZENDICT_TAG), Dumper=AiiDADumper
)
yaml.add_constructor(
    _PLUMPY_ATTRIBUTES_FROZENDICT_TAG, partial(mapping_constructor, AttributesFrozendict), Loader=AiiDALoader
)
yaml.add_constructor(_PLUMPY_BUNDLE, bundle_constructor, Loader=AiiDALoader)
yaml.add_constructor(_NODE_TAG, node_constructor, Loader=AiiDALoader)
yaml.add_constructor(_NODE_LINKS_MANAGER_TAG, node_links_manager_constructor, Loader=AiiDALoader)
yaml.add_constructor(_GROUP_TAG, group_constructor, Loader=AiiDALoader)
yaml.add_constructor(_COMPUTER_TAG, computer_constructor, Loader=AiiDALoader)
yaml.add_constructor(_ENUM_TAG, enum_constructor, Loader=AiiDALoader)
yaml.add_constructor(_DATACLASS_TAG, dataclass_constructor, Loader=AiiDALoader)


@overload
def serialize(data: Any, encoding: None = None) -> str: ...


@overload
def serialize(data: Any, encoding: str) -> bytes: ...


def serialize(data: Any, encoding: str | None = None) -> str | bytes:
    """Serialize the given data structure into a yaml dump.

    The function supports standard data containers such as maps and lists as well as AiiDA nodes which will be
    serialized into strings, before the whole data structure is dumped into a string using yaml.

    :param data: the general data to serialize
    :param encoding: optional encoding for the serialized string
    :return: string representation of the serialized data structure or byte array if specific encoding is specified
    """
    serialized: bytes | str

    if encoding is not None:
        serialized = yaml.dump(data, encoding=encoding, Dumper=AiiDADumper)
    else:
        serialized = yaml.dump(data, Dumper=AiiDADumper)

    return serialized


def deserialize_unsafe(serialized: str) -> Any:
    """Deserialize a yaml dump that represents a serialized data structure.

    .. note:: This function should not be used on untrusted input, since it is built upon `yaml.Loader` which is unsafe.

    :param serialized: a yaml serialized string representation
    :return: the deserialized data structure
    """
    return yaml.load(serialized, Loader=AiiDALoader)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi devel rabbitmq` commands."""

from __future__ import annotations

import re
import sys
import typing as t

import click
import wrapt

from aiida.cmdline.commands.cmd_devel import verdi_devel
from aiida.cmdline.params import arguments, options
from aiida.cmdline.utils import decorators, echo, echo_tabulate

if t.TYPE_CHECKING:
    import requests


@verdi_devel.group('rabbitmq')
def cmd_rabbitmq():
    """Commands to interact with RabbitMQ."""


@cmd_rabbitmq.group('queues')
def cmd_queues():
    """Commands to interact with RabbitMQ queues."""


@cmd_rabbitmq.group('tasks')
def cmd_tasks():
    """Commands related to process tasks."""


AVAILABLE_PROJECTORS = (
    'arguments',
    'auto_delete',
    'backing_queue_status',
    'consumer_utilisation',
    'consumers',
    'durable',
    'effective_policy_definition',
    'exclusive',
    'exclusive_consumer_tag',
    'garbage_collection',
    'head_message_timestamp',
    'idle_since',
    'memory',
    'message_bytes',
    'message_bytes_paged_out',
    'message_bytes_persistent',
    'message_bytes_ram',
    'message_bytes_ready',
    'message_bytes_unacknowledged',
    'messages',
    'messages_details',
    'messages_paged_out',
    'messages_persistent',
    'messages_ram',
    'messages_ready',
    'messages_ready_details',
    'messages_ready_ram',
    'messages_unacknowledged',
    'messages_unacknowledged_details',
    'messages_unacknowledged_ram',
    'name',
    'node',
    'operator_policy',
    'policy',
    'recoverable_slaves',
    'reductions',
    'reductions_details',
    'single_active_consumer_tag',
    'state',
    'type',
    'vhost',
)


def echo_response(response: 'requests.Response', exit_on_error: bool = True) -> None:
    """Echo the response of a request.

    :param response: The response to the request.
    :param exit_on_error: Boolean, if ``True``, call ``sys.exit`` with the status code of the response.
    """
    import requests

    try:
        response.raise_for_status()
    except requests.HTTPError:
        click.secho(f'[{response.status_code}] ', fg='red', bold=True, nl=False)
        click.secho(f'{response.reason}: {response.url}')
        if exit_on_error:
            sys.exit(response.status_code)
    else:
        click.secho(f'[{response.status_code}] ', fg='green', bold=True)


@wrapt.decorator
@click.pass_context
def with_client(ctx, wrapped, _, args, kwargs):
    """Decorate a function injecting a :class:`aiida.brokers.rabbitmq.client.RabbitmqManagementClient`."""
    from aiida.brokers.rabbitmq.client import RabbitmqManagementClient

    config = ctx.obj.profile.process_control_config
    client = RabbitmqManagementClient(
        username=config['broker_username'],
        password=config['broker_password'],
        hostname=config['broker_host'],
        virtual_host=config['broker_virtual_host'],
    )

    if not client.is_connected:
        echo.echo_critical(
            'Could not connect to the management API. Make sure RabbitMQ is running and the management plugin is '
            'installed using `sudo rabbitmq-plugins enable rabbitmq_management`. The API is served on port 15672, so '
            'if you are connecting to RabbitMQ running in a Docker container, make sure that port is exposed.'
        )

    kwargs['client'] = client
    return wrapped(*args, **kwargs)


@cmd_rabbitmq.command('server-properties')
@decorators.with_manager
def cmd_server_properties(manager):
    """List the server properties."""
    import yaml

    data = {}
    for key, value in manager.get_communicator().server_properties.items():
        data[key] = value.decode('utf-8') if isinstance(value, bytes) else value
    click.echo(yaml.dump(data, indent=4))


@cmd_queues.command('list')
@click.option(
    '-P',
    '--project',
    type=click.Choice(AVAILABLE_PROJECTORS),
    cls=options.MultipleValueOption,
    default=('name', 'messages', 'state'),
)
@options.RAW()
@click.option('-f', '--filter-name', type=str, help='Provide a regex pattern to filter queues based on their name. ')
@with_client
def cmd_queues_list(client, project, raw, filter_name):
    """List all queues."""
    response = client.request('queues')

    if not response.ok:
        echo_response(response)

    if filter_name and 'name' not in project:
        raise click.BadParameter('cannot use `--filter-name` when not projecting `name`.')

    if filter_name:
        try:
            re.match(filter_name, '')
        except re.error as exception:
            raise click.BadParameter(f'invalid regex pattern: {exception}', param_hint='`--filter-name`')

    queues = [queue for queue in response.json() if re.match(filter_name or '', queue['name'])]
    output = [
        list(map(lambda key, values=queue: values.get(key, ''), project))  # type: ignore[misc]
        for queue in queues
    ]

    if not output:
        echo.echo_report('No queues matched.')
        return

    headers = [name.capitalize() for name in project] if not raw else []
    tablefmt = None if not raw else 'plain'
    echo_tabulate(output, headers=headers, tablefmt=tablefmt)


@cmd_queues.command('create')
@click.argument('queues', nargs=-1)
@with_client
def cmd_queues_create(client, queues):
    """Create new queues."""
    for queue in queues:
        response = client.request('queues/{virtual_host}/{queue}', {'queue': queue}, method='PUT')
        click.secho(f'Create `{queue}`... ', nl=False)
        echo_response(response)


@cmd_queues.command('delete')
@click.argument('queues', nargs=-1)
@with_client
def cmd_queues_delete(client, queues):
    """Delete existing queues."""
    for queue in queues:
        params = {'if-empty': True, 'if-unused': True}
        response = client.request('queues/{virtual_host}/{queue}', {'queue': queue}, method='DELETE', params=params)
        click.secho(f'Delete `{queue}`... ', nl=False)
        echo_response(response, exit_on_error=False)


@cmd_tasks.command('list')
@decorators.with_broker
@decorators.only_if_daemon_not_running()
def cmd_tasks_list(broker):
    """List all active process tasks.

    This command prints a list of process pk's for which there is an active process task with RabbitMQ. Since tasks can
    only be seen when they are not currently with a daemon worker, this command can only be run when the daemon is not
    running.
    """
    from aiida.engine.processes.control import get_process_tasks

    for pk in get_process_tasks(broker):
        echo.echo(pk)


@cmd_tasks.command('analyze', deprecated='Use `verdi process repair` instead.')
@click.option('--fix', is_flag=True, help='Attempt to fix the inconsistencies if any are detected.')
@decorators.only_if_daemon_not_running()
@click.pass_context
def cmd_tasks_analyze(ctx, fix):
    """Perform analysis of process tasks.

    This command will perform a query of the database to find all "active" processes, meaning those that haven't yet
    reached a terminal state, and cross-references this with the active process tasks that are in the process queue of
    RabbitMQ. Any active process that does not have a corresponding process task can be considered a zombie, as it will
    never be picked up by a daemon worker to complete it and will effectively be "stuck". Any process task that does not
    correspond to an active process is useless and should be discarded. Finally, duplicate process tasks are also
    problematic and duplicates should be discarded.

    Use ``-v INFO`` to be more verbose and print more information.
    """
    from .cmd_process import process_repair

    ctx.invoke(process_repair, dry_run=not fix)


@cmd_tasks.command('revive')
@arguments.PROCESSES()
@options.FORCE()
@decorators.only_if_daemon_running(message='The daemon has to be running for this command to work.')
def cmd_tasks_revive(processes, force):
    """Revive processes that seem stuck and are no longer reachable.

    Warning: Use only as a last resort after you've gone through the checklist below.

    \b
        1. Does ``verdi status`` indicate that both daemon and RabbitMQ are running properly?
           If not, restart the daemon with ``verdi daemon restart`` and restart RabbitMQ.
        2. Try ``verdi process play <PID>``.
           If you receive a message that the process is no longer reachable,
           use ``verdi devel rabbitmq tasks revive <PID>``.

    Details: When RabbitMQ loses the process task before the process has completed, the process is never picked up by
    the daemon and will remain "stuck". ``verdi devel rabbitmq tasks revive`` recreates the task, which can lead to
    multiple instances of the task being executed and should thus be used with caution.
    """
    from aiida.engine.processes.control import revive_processes

    if not force:
        echo.echo_warning('This command should only be used if you are absolutely sure the process task was lost.')
        click.confirm(text='Do you want to continue?', abort=True)

    revive_processes(processes)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
""".. py:module::config
    :synopsis: Convenience class for configuration file option

The functions :func:`configuration_callback` and :func:`configuration_option` were directly taken from the repository
https://github.com/phha/click_config_file/blob/7b93a20b4c79458987fac116418859f30a16d82a/click_config_file.py with a
minor modification to ``configuration_callback`` to add a check for unknown parameters in the configuration file and
the default provider is changed to :func:`yaml_config_file_provider`.
"""

from __future__ import annotations

import functools
import os
import typing as t

import click

from .overridable import OverridableOption

__all__ = ('ConfigFileOption',)


def yaml_config_file_provider(handle, cmd_name):
    """Read yaml config file from file handle."""
    import yaml

    return yaml.safe_load(handle)


def configuration_callback(
    cmd_name: str | None,
    option_name: str,
    config_file_name: str,
    saved_callback: t.Callable[..., t.Any] | None,
    provider: t.Callable[..., t.Any],
    implicit: bool,
    ctx: click.Context,
    param: click.Parameter,
    value: t.Any,
):
    """Callback for reading the config file.

    Also takes care of calling user specified custom callback afterwards.

    :param cmd_name: The command name. This is used to determine the configuration directory.
    :param option_name: The name of the option. This is used for error messages.
    :param config_file_name: The name of the configuration file.
    :param saved_callback: User-specified callback to be called later.
    :param provider: A callable that parses the configuration file and returns a dictionary of the configuration
        parameters. Will be called as ``provider(file_path, cmd_name)``. Default: ``yaml_config_file_provider``.
    :param implicit: Whether a implicit value should be applied if no configuration option value was provided.
    :param ctx: ``click`` context.
    :param param: ``click`` parameters.
    :param value: Value passed to the parameter.
    """
    ctx.default_map = ctx.default_map or {}
    cmd_name = cmd_name or ctx.info_name

    if implicit:
        default_value = os.path.join(click.get_app_dir(cmd_name), config_file_name)  # type: ignore[arg-type]
        param.default = default_value
        value = value or default_value

    if value:
        try:
            config = provider(value, cmd_name)
        except Exception as exception:
            raise click.BadOptionUsage(option_name, f'Error reading configuration file: {exception}', ctx)

        valid_params = [param.name for param in ctx.command.params if param.name != option_name]
        specified_params = list(config.keys())
        unknown_params = set(specified_params).difference(set(valid_params))

        if unknown_params:
            raise click.BadParameter(
                f'Invalid configuration file, the following keys are not supported: {unknown_params}', ctx, param
            )

        ctx.default_map.update(config)

    return saved_callback(ctx, param, value) if saved_callback else value


def configuration_option(*param_decls, **attrs):
    """Adds configuration file support to a click application.

    This will create an option of type ``click.File`` expecting the path to a configuration file. When specified, it
    overwrites the default values for all other click arguments or options with the corresponding value from the
    configuration file. The default name of the option is ``--config``. By default, the configuration will be read from
    a configuration directory as determined by ``click.get_app_dir``. This decorator accepts the same arguments as
    ``click.option`` and ``click.Path``. In addition, the following keyword arguments are available:

    :param cmd_name: str
        The command name. This is used to determine the configuration directory. Default: ``ctx.info_name``.
    :param config_file_name: str
        The name of the configuration file. Default: ``config``.
    :param implicit: bool
        If ``True`` then implicitly create a value for the configuration option using the above parameters. If a
        configuration file exists in this path it will be applied even if no configuration option was suppplied as a
        CLI argument or environment variable. If ``False`` only apply a configuration file that has been explicitely
        specified. Default: ``False``.
    :param provider: callable
        A callable that parses the configuration file and returns a dictionary of the configuration parameters. Will be
        called as ``provider(file_path, cmd_name)``. Default: ``yaml_config_file_provider``.
    """
    param_decls = param_decls or ('--config',)
    option_name = param_decls[0]

    def decorator(func):
        attrs.setdefault('is_eager', True)
        attrs.setdefault('help', 'Read configuration from FILE.')
        attrs.setdefault('expose_value', False)
        implicit = attrs.pop('implicit', True)
        cmd_name = attrs.pop('cmd_name', None)
        config_file_name = attrs.pop('config_file_name', 'config')
        provider = attrs.pop('provider', yaml_config_file_provider)
        path_default_params = {
            'exists': False,
            'file_okay': True,
            'dir_okay': False,
            'writable': False,
            'readable': True,
            'resolve_path': False,
        }
        path_params = {k: attrs.pop(k, v) for k, v in path_default_params.items()}
        attrs['type'] = attrs.get('type', click.Path(**path_params))
        saved_callback = attrs.pop('callback', None)
        partial_callback = functools.partial(
            configuration_callback, cmd_name, option_name, config_file_name, saved_callback, provider, implicit
        )
        attrs['callback'] = partial_callback
        return click.option(*param_decls, **attrs)(func)

    return decorator


class ConfigFileOption(OverridableOption):
    """Reusable option that reads a configuration file containing values for other command parameters.

    Example::

        CONFIG_FILE = ConfigFileOption('--config', help='A configuration file')

        @click.command()
        @click.option('computer_name')
        @CONFIG_FILE(help='Configuration file for computer_setup')
        def computer_setup(computer_name):
            click.echo(f"Setting up computer {computername}")

        computer_setup --config config.yml

    with config.yml::

        ---
        computer_name: computer1

    """

    def __init__(self, *args, **kwargs):
        """Store the default args and kwargs.

        :param args: default arguments to be used for the option
        :param kwargs: default keyword arguments to be used that can be overridden in the call
        """
        kwargs.update({'provider': yaml_config_file_provider, 'implicit': False})
        super().__init__(*args, **kwargs)

    def __call__(self, **kwargs):
        """Override the stored kwargs, (ignoring args as we do not allow option name changes) and return the option.

        :param kwargs: keyword arguments that will override those set in the construction
        :return: click_config_file.configuration_option constructed with args and kwargs defined during construction
            and call of this instance
        """
        kw_copy = self.kwargs.copy()
        kw_copy.update(kwargs)

        return configuration_option(*self.args, **kw_copy)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data core.bands` command."""

import click

from aiida.cmdline.commands.cmd_data import cmd_show, verdi_data
from aiida.cmdline.commands.cmd_data.cmd_export import data_export
from aiida.cmdline.commands.cmd_data.cmd_list import list_options
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo
from aiida.common.utils import Prettifier

LIST_PROJECT_HEADERS = ['ID', 'Formula', 'Ctime', 'Label']
EXPORT_FORMATS = [
    'agr',
    'agr_batch',
    'dat_blocks',
    'dat_multicolumn',
    'gnuplot',
    'json',
    'mpl_pdf',
    'mpl_png',
    'mpl_singlefile',
    'mpl_withjson',
]
VISUALIZATION_FORMATS = ['xmgrace']


@verdi_data.group('core.bands')
def bands():
    """Manipulate BandsData objects (band structures)."""


@bands.command('list')
@decorators.with_dbenv()
@list_options
@options.WITH_ELEMENTS()
@options.WITH_ELEMENTS_EXCLUSIVE()
@options.FORMULA_MODE()
def bands_list(elements, elements_exclusive, raw, formula_mode, past_days, groups, all_users):
    """List BandsData objects."""
    from argparse import Namespace

    from tabulate import tabulate

    from aiida.orm.nodes.data.array.bands import get_bands_and_parents_structure

    args = Namespace()
    args.element = elements
    args.element_only = elements_exclusive
    args.formulamode = formula_mode
    args.past_days = past_days
    args.group_name = None
    if groups is not None:
        args.group_pk = [group.pk for group in groups]
    else:
        args.group_pk = None
    args.all_users = all_users

    entry_list = get_bands_and_parents_structure(args)

    counter = 0
    bands_list_data = []
    if not raw:
        bands_list_data.append(LIST_PROJECT_HEADERS)
    for entry in entry_list:
        for i, value in enumerate(entry):
            if isinstance(value, list):
                entry[i] = ','.join(value)
        for i in range(len(entry), len(LIST_PROJECT_HEADERS)):
            entry.append(None)
        counter += 1
    bands_list_data.extend(entry_list)
    if raw:
        echo.echo(tabulate(bands_list_data, tablefmt='plain'))
    else:
        echo.echo(tabulate(bands_list_data, headers='firstrow'))
        echo.echo(f'\nTotal results: {counter}\n')


@bands.command('show')
@arguments.DATA(type=types.DataParamType(sub_classes=('aiida.data:core.array.bands',)))
@options.VISUALIZATION_FORMAT(type=click.Choice(VISUALIZATION_FORMATS), default='xmgrace')
@decorators.with_dbenv()
def bands_show(data, fmt):
    """Visualize BandsData objects."""
    try:
        show_function = getattr(cmd_show, f'_show_{fmt}')
    except AttributeError:
        echo.echo_critical(f'visualization format {fmt} is not supported')

    show_function(fmt, data)


@bands.command('export')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:core.array.bands',)))
@options.EXPORT_FORMAT(type=click.Choice(EXPORT_FORMATS), default='json')
@click.option(
    '--y-min-lim',
    type=click.FLOAT,
    default=None,
    help='The minimum value for the y axis.' ' Default: minimum of all bands',
)
@click.option(
    '--y-max-lim',
    type=click.FLOAT,
    default=None,
    help='The maximum value for the y axis.' ' Default: maximum of all bands',
)
@click.option(
    '-o',
    '--output',
    type=click.STRING,
    default=None,
    help='If present, store the output directly on a file '
    'with the given name. It is essential to use this option '
    'if more than one file needs to be created.',
)
@options.FORCE(help='If passed, overwrite files without checking.')
@click.option(
    '--prettify-format',
    default=None,
    type=click.Choice(Prettifier.get_prettifiers()),
    help='The style of labels for the prettifier',
)
@decorators.with_dbenv()
def bands_export(fmt, y_min_lim, y_max_lim, output, force, prettify_format, datum):
    """Export BandsData objects."""
    args = {}
    if y_min_lim is not None:
        args['y_min_lim'] = y_min_lim
    if y_max_lim is not None:
        args['y_max_lim'] = y_max_lim
    if prettify_format is not None:
        args['prettify_format'] = prettify_format

    data_export(datum, output, fmt, other_args=args, overwrite=force)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Mixin class for SQL implementations of ``extras``."""

from typing import Any, Dict, Iterable, Tuple

from aiida.orm.implementation.utils import clean_value, validate_attribute_extra_key


class ExtrasMixin:
    """Mixin class for SQL implementations of ``extras``."""

    model: Any
    bare_model: Any
    is_stored: bool

    @property
    def extras(self) -> Dict[str, Any]:
        return self.model.extras

    def get_extra(self, key: str) -> Any:
        try:
            return self.model.extras[key]
        except KeyError as exception:
            raise AttributeError(f'extra `{exception}` does not exist') from exception

    def set_extra(self, key: str, value: Any) -> None:
        validate_attribute_extra_key(key)

        if self.is_stored:
            value = clean_value(value)

        self.model.extras[key] = value
        self._flush_if_stored({'extras'})

    def set_extra_many(self, extras: Dict[str, Any]) -> None:
        for key in extras:
            validate_attribute_extra_key(key)

        if self.is_stored:
            extras = {key: clean_value(value) for key, value in extras.items()}

        for key, value in extras.items():
            self.bare_model.extras[key] = value

        self._flush_if_stored({'extras'})

    def reset_extras(self, extras: Dict[str, Any]) -> None:
        for key in extras:
            validate_attribute_extra_key(key)

        if self.is_stored:
            extras = clean_value(extras)

        self.bare_model.extras = extras
        self._flush_if_stored({'extras'})

    def delete_extra(self, key: str) -> None:
        try:
            self.model.extras.pop(key)
        except KeyError as exception:
            raise AttributeError(f'extra `{exception}` does not exist') from exception
        else:
            self._flush_if_stored({'extras'})

    def delete_extra_many(self, keys: Iterable[str]) -> None:
        non_existing_keys = [key for key in keys if key not in self.model.extras]

        if non_existing_keys:
            raise AttributeError(f"extras `{', '.join(non_existing_keys)}` do not exist")

        for key in keys:
            self.bare_model.extras.pop(key)

        self._flush_if_stored({'extras'})

    def clear_extras(self) -> None:
        self.model.extras = {}
        self._flush_if_stored({'extras'})

    def extras_items(self) -> Iterable[Tuple[str, Any]]:
        for key, value in self.model.extras.items():
            yield key, value

    def extras_keys(self) -> Iterable[str]:
        for key in self.model.extras.keys():
            yield key



#!/usr/bin/env python
"""Pre-commit hook to add ``__all__`` imports to ``__init__`` files."""

import ast
import sys
from collections import Counter
from pathlib import Path
from pprint import pprint
from typing import Dict, List, Optional, Tuple


def parse_all(folder_path: str) -> Tuple[dict, dict]:
    """Walk through all files in folder, and parse the ``__all__`` variable.

    :return: (all dict, dict of unparsable)
    """
    folder_path = Path(folder_path)
    all_dict = {}
    bad_all = {}

    for path in folder_path.glob('**/*.py'):
        # skip module files
        if path.name == '__init__.py':
            continue

        # parse the file
        parsed = ast.parse(path.read_text(encoding='utf8'))

        # find __all__ assignment
        all_token = None
        for token in parsed.body:
            if not isinstance(token, ast.Assign):
                continue
            if token.targets and getattr(token.targets[0], 'id', '') == '__all__':
                all_token = token
                break

        if all_token is None:
            bad_all.setdefault('missing', []).append(str(path.relative_to(folder_path)))
            continue

        if not isinstance(all_token.value, (ast.List, ast.Tuple)):
            bad_all.setdefault('value not list/tuple', []).append(str(path.relative_to(folder_path)))
            continue

        if not all(isinstance(el, ast.Str) for el in all_token.value.elts):
            bad_all.setdefault('child not strings', []).append(str(path.relative_to(folder_path)))
            continue

        names = [n.s for n in all_token.value.elts]
        if not names:
            continue

        path_dict = all_dict
        for part in path.parent.relative_to(folder_path).parts:
            path_dict = path_dict.setdefault(part, {})
        path_dict.setdefault(path.name[:-3], {})['__all__'] = names

    return all_dict, bad_all


def gather_all(
    cur_path: List[str], cur_dict: dict, skip_children: dict, all_list: Optional[List[str]] = None
) -> List[str]:
    """Recursively gather __all__ names."""
    all_list = [] if all_list is None else all_list
    for key, val in cur_dict.items():
        if key == '__all__':
            all_list.extend(val)
        elif key not in skip_children.get('/'.join(cur_path), []):
            gather_all(cur_path + [key], val, skip_children, all_list)
    return all_list


def write_inits(folder_path: str, all_dict: dict, skip_children: Dict[str, List[str]]) -> Dict[str, List[str]]:
    """Write __init__.py files for all subfolders.

    :return: folders with non-unique imports
    """
    folder_path = Path(folder_path)
    non_unique = {}
    for path in folder_path.glob('**/__init__.py'):
        if path.parent == folder_path:
            # skip top level __init__.py
            continue

        rel_path = path.parent.relative_to(folder_path).as_posix()

        # get sub_dict for this folder
        path_all_dict = all_dict
        mod_path = path.parent.relative_to(folder_path).parts
        try:
            for part in mod_path:
                path_all_dict = path_all_dict[part]
        except KeyError:
            # there is nothing to import
            continue

        if '*' in skip_children.get(rel_path, []):
            path_all_dict = {}
            alls = []
            auto_content = ['', '# AUTO-GENERATED', '', '__all__ = ()', '']
        else:
            path_all_dict = {
                key: val for key, val in path_all_dict.items() if key not in skip_children.get(rel_path, [])
            }
            alls = gather_all(list(mod_path), path_all_dict, skip_children)

            # check for non-unique imports
            if len(alls + list(path_all_dict)) != len(set(alls + list(path_all_dict))):
                non_unique[rel_path] = [k for k, v in Counter(alls + list(path_all_dict)).items() if v > 1]

            auto_content = (
                ['', '# AUTO-GENERATED']
                + ['', '# fmt: off', '']
                + [f'from .{mod} import *' for mod in sorted(path_all_dict.keys())]
                + ['', '__all__ = (']
                + [f'    {a!r},' for a in sorted(set(alls))]
                + [')', '', '# fmt: on', '']
            )

        start_content = []
        end_content = []
        in_docstring = in_end_content = False
        in_start_content = True
        for line in path.read_text(encoding='utf8').splitlines():
            if not in_start_content and line.startswith('# END AUTO-GENERATED'):
                in_end_content = True
            if in_end_content:
                end_content.append(line)
                continue
            # only use initial comments and docstring
            if not (line.startswith('#') or line.startswith('"""') or in_docstring):
                in_start_content = False
                continue
            if line.startswith('"""'):
                if (not in_docstring) and not (line.endswith('"""') and not line.strip() == '"""'):
                    in_docstring = True
                else:
                    in_docstring = False
            if in_start_content:
                start_content.append(line)

        new_content = start_content + auto_content + end_content

        path.write_text('\n'.join(new_content).rstrip() + '\n', encoding='utf8')

    return non_unique


if __name__ == '__main__':
    _folder = Path(__file__).parent.parent.joinpath('aiida')
    _skip = {
        # skipped since some arguments and options share the same name
        'cmdline/params': ['arguments', 'options'],
        # skipped since this is for testing only not general use
        'manage': ['tests'],
        # skipped since we don't want to expose the implementation at the top-level
        'storage': ['psql_dos', 'sqlite_zip', 'sqlite_temp'],
        'orm': ['implementation'],
        # skip all since the module requires extra requirements
        'restapi': ['*'],
        # keep at aiida.tools.archive level
        'tools': ['archive'],
    }
    _all_dict, _bad_all = parse_all(_folder)
    _non_unique = write_inits(_folder, _all_dict, _skip)
    _bad_all.pop('missing', '')  # allow missing __all__
    if _bad_all:
        print('unparsable __all__:')
        pprint(_bad_all)
    if _non_unique:
        print('non-unique imports:')
        pprint(_non_unique)
    if _bad_all or _non_unique:
        sys.exit(1)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Drop `db_dbcomputer.transport_params`

This is similar to migration 07fac78e6209

Revision ID: django_0036
Revises: django_0035

"""

from alembic import op

revision = 'django_0036'
down_revision = 'django_0035'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.drop_column('db_dbcomputer', 'transport_params')


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0036.')
